import os

import flask
from flask import current_app
from flask import g

from canaille.app import DOCUMENTATION_URL

try:
    import flask_themer
except ImportError:
    flask_themer = None


def render_template(*args, **kwargs):
    if flask_themer and current_app.config["CANAILLE"]["THEME"]:
        return flask_themer.render_template(*args, **kwargs)

    return flask.render_template(*args, **kwargs)


def setup_themer(app):
    theme_path = app.config["CANAILLE"]["THEME"]

    if not theme_path:
        with app.app_context():

            @current_app.context_processor
            def global_processor():
                return {"theme": lambda theme: theme}

        return

    theme_path = os.path.abspath(theme_path)
    themes_dir, theme_name = theme_path.rsplit("/", 1)

    themer = flask_themer.Themer(
        app, loaders=[flask_themer.FileSystemThemeLoader(themes_dir)]
    )

    @themer.current_theme_loader
    def get_current_theme():
        return theme_name


def setup_jinja(app):
    from canaille.app.forms import password_strength_calculator
    from canaille.app.models import MODELS
    from canaille.backends import Backend

    app.jinja_env.filters["len"] = len
    app.jinja_env.filters["password_strength"] = password_strength_calculator

    def create_model_filter(model_class):
        """Create a filter function for a specific model class."""

        def get_by_id(object_id):
            return Backend.instance.get(model_class, object_id) if object_id else None

        return get_by_id

    for model_class in MODELS.values():
        app.jinja_env.filters[model_class.__name__] = create_model_filter(model_class)

    app.jinja_env.policies["ext.i18n.trimmed"] = True

    @app.context_processor
    def global_processor():
        from canaille.app.flask import request_is_boosted
        from canaille.app.flask import request_is_partial

        return {
            "debug": app.debug or app.config.get("TESTING", False),
            "documentation_url": DOCUMENTATION_URL,
            "logo_url": app.config["CANAILLE"]["LOGO"],
            "favicon_url": app.config["CANAILLE"]["FAVICON"]
            or app.config["CANAILLE"]["LOGO"],
            "website_name": app.config["CANAILLE"]["NAME"],
            "user": g.get("session") and g.session.user,
            "menu": True,
            "request_is_boosted": request_is_boosted(),
            "request_is_partial": request_is_partial(),
            "features": app.features,
            "no_secret_key": app.no_secret_key,
        }



from canaille.app import models

# TODO: this test files tests generic features about forms,
# we should rewrite the tests without using specific OIDC
# forms.


def test_fieldlist_add(testclient, logged_admin, backend):
    assert not backend.query(models.Client)

    res = testclient.get("/admin/client/add")
    assert "redirect_uris-1" not in res.form.fields

    data = {
        "client_name": "foobar",
        "client_uri": "https://foobar.test",
        "redirect_uris-0": "https://foobar.test/callback",
        "grant_types": ["password", "authorization_code"],
        "response_types": ["code", "token"],
        "token_endpoint_auth_method": "none",
    }
    for k, v in data.items():
        res.form[k].force_value(v)

    res = res.form.submit(status=200, name="fieldlist_add", value="redirect_uris-0")
    assert not backend.query(models.Client)

    data["redirect_uris-1"] = "https://foobar.test/callback2"
    for k, v in data.items():
        res.form[k].force_value(v)

    res = res.form.submit(status=302, name="action", value="edit")
    res = res.follow(status=200)

    client_id = res.forms["readonly"]["client_id"].value
    client = backend.get(models.Client, client_id=client_id)

    assert client.redirect_uris == [
        "https://foobar.test/callback",
        "https://foobar.test/callback2",
    ]

    backend.delete(client)


def test_fieldlist_delete(testclient, logged_admin, backend):
    assert not backend.query(models.Client)
    res = testclient.get("/admin/client/add")

    data = {
        "client_name": "foobar",
        "client_uri": "https://foobar.test",
        "redirect_uris-0": "https://foobar.test/callback1",
        "grant_types": ["password", "authorization_code"],
        "response_types": ["code", "token"],
        "token_endpoint_auth_method": "none",
    }
    for k, v in data.items():
        res.form[k].force_value(v)
    res = res.form.submit(status=200, name="fieldlist_add", value="redirect_uris-0")

    res.form["redirect_uris-1"] = "https://foobar.test/callback2"
    res = res.form.submit(status=200, name="fieldlist_remove", value="redirect_uris-1")
    assert not backend.query(models.Client)
    assert "redirect_uris-1" not in res.form.fields

    res = res.form.submit(status=302, name="action", value="edit")
    res = res.follow(status=200)

    client_id = res.forms["readonly"]["client_id"].value
    client = backend.get(models.Client, client_id=client_id)

    assert client.redirect_uris == [
        "https://foobar.test/callback1",
    ]

    backend.delete(client)


def test_fieldlist_add_invalid_field(testclient, logged_admin):
    res = testclient.get("/admin/client/add")
    data = {
        "csrf_token": res.form["csrf_token"].value,
        "client_name": "foobar",
        "client_uri": "https://foobar.test",
        "redirect_uris-0": "https://foobar.test/callback",
        "grant_types": ["password", "authorization_code"],
        "response_types": ["code", "token"],
        "token_endpoint_auth_method": "none",
        "fieldlist_add": "invalid",
    }
    testclient.post("/admin/client/add", data, status=400)


def test_fieldlist_delete_invalid_field(testclient, logged_admin, backend):
    assert not backend.query(models.Client)
    res = testclient.get("/admin/client/add")

    data = {
        "csrf_token": res.form["csrf_token"].value,
        "client_name": "foobar",
        "client_uri": "https://foobar.test",
        "redirect_uris-0": "https://foobar.test/callback1",
        "redirect_uris-1": "https://foobar.test/callback2",
        "grant_types": ["password", "authorization_code"],
        "response_types": ["code", "token"],
        "token_endpoint_auth_method": "none",
        "fieldlist_remove": "invalid",
    }
    testclient.post("/admin/client/add", data, status=400)


def test_fieldlist_duplicate_value(testclient, logged_admin, client):
    res = testclient.get("/admin/client/add")
    data = {
        "client_name": "foobar",
        "client_uri": "https://foobar.test",
        "redirect_uris-0": "https://foobar.test/samecallback",
        "grant_types": ["password", "authorization_code"],
        "response_types": ["code", "token"],
        "token_endpoint_auth_method": "none",
    }
    for k, v in data.items():
        res.form[k].force_value(v)
    res = res.form.submit(status=200, name="fieldlist_add", value="redirect_uris-0")
    res.form["redirect_uris-1"] = "https://foobar.test/samecallback"
    res = res.form.submit(status=200, name="action", value="edit")
    res.mustcontain("This value is a duplicate")


def test_fieldlist_empty_value(testclient, logged_admin, backend):
    res = testclient.get("/admin/client/add")
    data = {
        "client_name": "foobar",
        "client_uri": "https://foobar.test",
        "redirect_uris-0": "https://foobar.test/samecallback",
        "post_logout_redirect_uris-0": "https://foobar.test/callback1",
        "grant_types": ["password", "authorization_code"],
        "response_types": ["code", "token"],
        "token_endpoint_auth_method": "none",
    }
    for k, v in data.items():
        res.form[k].force_value(v)
    res = res.form.submit(
        status=200, name="fieldlist_add", value="post_logout_redirect_uris-0"
    )
    res.form.submit(status=302, name="action", value="edit")
    client = backend.get(models.Client)
    backend.delete(client)


def test_fieldlist_add_field_htmx(testclient, logged_admin):
    res = testclient.get("/admin/client/add")
    data = {
        "csrf_token": res.form["csrf_token"].value,
        "client_name": "foobar",
        "client_uri": "https://foobar.test",
        "redirect_uris-0": "https://foobar.test/callback",
        "grant_types": ["password", "authorization_code"],
        "response_types": ["code", "token"],
        "token_endpoint_auth_method": "none",
        "fieldlist_add": "redirect_uris-0",
    }
    response = testclient.post(
        "/admin/client/add",
        data,
        headers={
            "HX-Request": "true",
            "HX-Trigger-Name": "listfield_add",
        },
    )
    assert 'name="redirect_uris-0' in response.text
    assert 'name="redirect_uris-1' in response.text


def test_fieldlist_add_field_htmx_validation(testclient, logged_admin):
    res = testclient.get("/admin/client/add")
    data = {
        "csrf_token": res.form["csrf_token"].value,
        "client_name": "foobar",
        "client_uri": "https://foobar.test",
        "redirect_uris-0": "not-a-valid-uri",
        "grant_types": ["password", "authorization_code"],
        "response_types": ["code", "token"],
        "token_endpoint_auth_method": "none",
        "fieldlist_add": "redirect_uris-0",
    }
    response = testclient.post(
        "/admin/client/add",
        data,
        headers={
            "HX-Request": "true",
            "HX-Trigger-Name": "listfield_add",
        },
    )
    assert 'name="redirect_uris-0' in response.text
    assert 'name="redirect_uris-1' in response.text
    assert "This is not a valid URL" in response.text


def test_fieldlist_remove_field_htmx(testclient, logged_admin):
    res = testclient.get("/admin/client/add")
    data = {
        "csrf_token": res.form["csrf_token"].value,
        "client_name": "foobar",
        "client_uri": "https://foobar.test",
        "redirect_uris-0": "https://foobar.test/callback1",
        "redirect_uris-1": "https://foobar.test/callback2",
        "grant_types": ["password", "authorization_code"],
        "response_types": ["code", "token"],
        "token_endpoint_auth_method": "none",
        "fieldlist_remove": "redirect_uris-1",
    }
    response = testclient.post(
        "/admin/client/add",
        data,
        headers={
            "HX-Request": "true",
            "HX-Trigger-Name": "listfield_remove",
        },
    )
    assert 'name="redirect_uris-0' in response.text
    assert 'name="redirect_uris-1' not in response.text


def test_fieldlist_inline_validation(testclient, logged_admin):
    res = testclient.get("/admin/client/add")
    data = {
        "csrf_token": res.form["csrf_token"].value,
        "client_name": "foobar",
        "client_uri": "https://foobar.test",
        "redirect_uris-0": "invalid-url",
        "redirect_uris-1": "https://foobar.test/callback2",
        "grant_types": ["password", "authorization_code"],
        "response_types": ["code", "token"],
        "token_endpoint_auth_method": "none",
    }
    response = testclient.post(
        "/admin/client/add",
        data,
        headers={
            "HX-Request": "true",
            "HX-Trigger-Name": "redirect_uris-0",
        },
    )
    assert 'name="redirect_uris-0' in response.text
    assert 'name="redirect_uris-1' not in response.text
    assert "This is not a valid URL" in response.text



import json
import warnings
from datetime import datetime

from joserfc.jwk import KeySet

from canaille.app import models


def test_get(testclient, backend, client, user, client_jwk):
    key_set = KeySet([client_jwk]).as_dict(private=False)
    assert not testclient.app.config["CANAILLE_OIDC"].get(
        "DYNAMIC_CLIENT_REGISTRATION_OPEN"
    )
    testclient.app.config["CANAILLE_OIDC"]["DYNAMIC_CLIENT_REGISTRATION_TOKENS"] = [
        "static-token"
    ]

    headers = {"Authorization": "Bearer static-token"}
    res = testclient.get(
        f"/oauth/register/{client.client_id}", headers=headers, status=200
    )
    assert res.json == {
        "application_type": None,
        "client_id": client.client_id,
        "client_secret": client.client_secret,
        "client_id_issued_at": int(datetime.timestamp(client.client_id_issued_at)),
        "client_secret_expires_at": 0,
        "default_acr_values": [],
        "default_max_age": None,
        "redirect_uris": [
            "https://client.test/redirect1",
            "https://client.test/redirect2",
        ],
        "registration_access_token": "static-token",
        "registration_client_uri": f"http://canaille.test/oauth/register/{client.client_id}",
        "token_endpoint_auth_method": "client_secret_basic",
        "grant_types": [
            "password",
            "authorization_code",
            "implicit",
            "hybrid",
            "refresh_token",
            "client_credentials",
            "urn:ietf:params:oauth:grant-type:jwt-bearer",
        ],
        "id_token_encrypted_response_alg": None,
        "id_token_encrypted_response_enc": None,
        "id_token_signed_response_alg": None,
        "initiate_login_uri": None,
        "response_types": ["code", "token", "id_token"],
        "client_name": "Some client",
        "client_uri": "https://client.test",
        "logo_uri": "https://client.test/logo.webp",
        "request_object_encryption_alg": None,
        "request_object_encryption_enc": None,
        "request_object_signing_alg": None,
        "request_uris": [],
        "require_auth_time": None,
        "scope": "openid email profile groups address phone",
        "sector_identifier_uri": None,
        "subject_type": None,
        "contacts": ["contact@mydomain.test"],
        "token_endpoint_auth_signing_alg": None,
        "tos_uri": "https://client.test/tos",
        "policy_uri": "https://client.test/policy",
        "jwks": json.dumps(key_set),
        "jwks_uri": None,
        "software_id": None,
        "software_version": None,
        "userinfo_encrypted_response_alg": None,
        "userinfo_encrypted_response_enc": None,
        "userinfo_signed_response_alg": None,
        "require_signed_request_object": None,
    }


def test_update(testclient, backend, client, user, client_jwk):
    key_set = KeySet([client_jwk]).as_dict(private=False)
    assert not testclient.app.config["CANAILLE_OIDC"].get(
        "DYNAMIC_CLIENT_REGISTRATION_OPEN"
    )
    testclient.app.config["CANAILLE_OIDC"]["DYNAMIC_CLIENT_REGISTRATION_TOKENS"] = [
        "static-token"
    ]

    assert client.redirect_uris != ["https://newname.example.test/callback"]
    assert client.token_endpoint_auth_method != "none"
    assert client.grant_types != ["refresh_token"]
    assert client.response_types != ["code"]
    assert client.client_name != "new name"
    assert client.client_uri != "https://newname.example.test"
    assert client.logo_uri != "https://newname.example.test/logo.webp"
    assert client.scope != ["openid", "profile", "email"]
    assert client.contacts != ["newcontact@example.test"]
    assert client.tos_uri != "https://newname.example.test/tos"
    assert client.policy_uri != "https://newname.example.test/policy"
    assert client.jwks_uri != "https://newname.example.test/my_public_keys.jwks"
    assert client.software_id != "new_software_id"
    assert client.software_version != "3.14"

    payload = {
        "client_id": client.client_id,
        "redirect_uris": ["https://newname.example.test/callback"],
        "token_endpoint_auth_method": "none",
        "grant_types": ["refresh_token"],
        "response_types": ["code"],
        "client_name": "new name",
        "client_uri": "https://newname.example.test",
        "logo_uri": "https://newname.example.test/logo.webp",
        "scope": "openid profile email",
        "contacts": ["newcontact@example.test"],
        "tos_uri": "https://newname.example.test/tos",
        "policy_uri": "https://newname.example.test/policy",
        "jwks_uri": "https://newname.example.test/my_public_keys.jwks",
        "require_auth_time": True,
        "software_id": "new_software_id",
        "software_version": "3.14",
    }

    headers = {"Authorization": "Bearer static-token"}
    res = testclient.put_json(
        f"/oauth/register/{client.client_id}", payload, headers=headers, status=200
    )
    client = backend.get(models.Client, client_id=res.json["client_id"])

    assert res.json == {
        "application_type": "web",
        "client_id": client.client_id,
        "client_secret": client.client_secret,
        "client_id_issued_at": int(datetime.timestamp(client.client_id_issued_at)),
        "client_secret_expires_at": 0,
        "default_acr_values": [],
        "default_max_age": None,
        "id_token_encrypted_response_alg": None,
        "id_token_encrypted_response_enc": None,
        "id_token_signed_response_alg": "RS256",
        "initiate_login_uri": None,
        "redirect_uris": ["https://newname.example.test/callback"],
        "registration_access_token": "static-token",
        "registration_client_uri": f"http://canaille.test/oauth/register/{client.client_id}",
        "request_object_encryption_alg": None,
        "request_object_encryption_enc": None,
        "request_object_signing_alg": None,
        "request_uris": [],
        "require_auth_time": True,
        "token_endpoint_auth_method": "none",
        "grant_types": ["refresh_token"],
        "response_types": ["code"],
        "client_name": "new name",
        "client_uri": "https://newname.example.test",
        "logo_uri": "https://newname.example.test/logo.webp",
        "scope": "openid profile email",
        "sector_identifier_uri": None,
        "subject_type": None,
        "contacts": ["newcontact@example.test"],
        "token_endpoint_auth_signing_alg": None,
        "tos_uri": "https://newname.example.test/tos",
        "policy_uri": "https://newname.example.test/policy",
        "jwks": json.dumps(key_set),
        "jwks_uri": "https://newname.example.test/my_public_keys.jwks",
        "software_id": "new_software_id",
        "software_version": "3.14",
        "userinfo_encrypted_response_alg": None,
        "userinfo_encrypted_response_enc": None,
        "userinfo_signed_response_alg": None,
        "require_signed_request_object": False,
    }

    assert client.redirect_uris == ["https://newname.example.test/callback"]
    assert client.token_endpoint_auth_method == "none"
    assert client.grant_types == ["refresh_token"]
    assert client.response_types == ["code"]
    assert client.client_name == "new name"
    assert client.client_uri == "https://newname.example.test"
    assert client.logo_uri == "https://newname.example.test/logo.webp"
    assert client.scope == ["openid", "profile", "email"]
    assert client.contacts == ["newcontact@example.test"]
    assert client.tos_uri == "https://newname.example.test/tos"
    assert client.policy_uri == "https://newname.example.test/policy"
    assert client.jwks_uri == "https://newname.example.test/my_public_keys.jwks"
    assert client.software_id == "new_software_id"
    assert client.software_version == "3.14"


def test_delete(testclient, backend, user):
    assert not testclient.app.config["CANAILLE_OIDC"].get(
        "DYNAMIC_CLIENT_REGISTRATION_OPEN"
    )
    testclient.app.config["CANAILLE_OIDC"]["DYNAMIC_CLIENT_REGISTRATION_TOKENS"] = [
        "static-token"
    ]

    client = models.Client(client_id="foobar", client_name="Some client")
    backend.save(client)

    headers = {"Authorization": "Bearer static-token"}
    with warnings.catch_warnings(record=True):
        testclient.delete(
            f"/oauth/register/{client.client_id}", headers=headers, status=204
        )
    assert not backend.get(models.Client, client_id=client.client_id)


def test_invalid_client(testclient, backend, user):
    assert not testclient.app.config["CANAILLE_OIDC"].get(
        "DYNAMIC_CLIENT_REGISTRATION_OPEN"
    )
    testclient.app.config["CANAILLE_OIDC"]["DYNAMIC_CLIENT_REGISTRATION_TOKENS"] = [
        "static-token"
    ]

    payload = {
        "client_id": "invalid-client-id",
        "redirect_uris": ["https://newname.example.test/callback"],
    }

    headers = {"Authorization": "Bearer static-token"}
    res = testclient.put_json(
        "/oauth/register/invalid-client-id", payload, headers=headers, status=401
    )
    assert res.json == {
        "error": "invalid_client",
        "error_description": "The client does not exist on this server.",
    }



import datetime
from dataclasses import dataclass

from flask import current_app
from flask import g
from flask import request
from flask import session
from itsdangerous import BadSignature
from itsdangerous.url_safe import URLSafeSerializer

from canaille.app import models
from canaille.core.models import User

USER_SESSION = "sessions"
LOGIN_HISTORY = "login_history"
LOGIN_HISTORY_COOKIE = "canaille_login_history"
MAX_LOGIN_HISTORY = 6
SESSION_LIFETIME = datetime.timedelta(days=365)


@dataclass
class UserSession:
    user: User | None = None
    last_login_datetime: datetime.datetime | None = None

    @classmethod
    def deserialize(cls, payload):
        user = current_app.backend.instance.get(models.User, payload.get("user"))
        user_is_locked = (
            user and current_app.backend.has_account_lockability() and user.locked
        )
        if not user or user_is_locked:
            return None

        return UserSession(
            user=user,
            last_login_datetime=datetime.datetime.fromisoformat(
                payload["last_login_datetime"]
            )
            if payload.get("last_login_datetime")
            else None,
        )

    def serialize(self):
        return {
            "user": self.user.id,
            "last_login_datetime": self.last_login_datetime.isoformat(),
        }


def current_user_session():
    for payload in session.get(USER_SESSION, [])[::-1]:
        if obj := UserSession.deserialize(payload):
            return obj

        session[USER_SESSION].remove(payload)

    if USER_SESSION in session and not session[USER_SESSION]:
        del session[USER_SESSION]

    return None


def save_user_session():
    session[USER_SESSION][-1] = g.session.serialize()


def login_user(user, remember: bool = True):
    """Open a session for the user.

    :param user: The user to log in
    :param remember: Whether to create a permanent session and add to login history
    """
    now = datetime.datetime.now(datetime.timezone.utc)
    obj = UserSession(user=user, last_login_datetime=now)
    g.session = obj
    try:
        previous = (
            session[USER_SESSION]
            if isinstance(session[USER_SESSION], list)
            else [session[USER_SESSION]]
        )
        session[USER_SESSION] = previous + [obj.serialize()]
    except KeyError:
        session[USER_SESSION] = [obj.serialize()]

    session.permanent = remember
    if remember:
        current_app.permanent_session_lifetime = SESSION_LIFETIME
        add_to_login_history(user.user_name)


def logout_user():
    """Close the user session."""
    try:
        session[USER_SESSION].pop()
        if not session[USER_SESSION]:
            del session[USER_SESSION]
    except (IndexError, KeyError):
        pass

    del g.session


def _get_login_history_from_cookie():
    """Get login history from secure cookie."""
    cookie_value = request.cookies.get(LOGIN_HISTORY_COOKIE)
    if not cookie_value:
        return []

    try:
        serializer = URLSafeSerializer(current_app.secret_key)
        data = serializer.loads(cookie_value)
        return data if isinstance(data, list) else []
    except (BadSignature, ValueError, TypeError):
        return []


def _set_login_history_cookie(login_history, response):
    """Set login history in secure cookie."""
    serializer = URLSafeSerializer(current_app.secret_key)
    cookie_value = serializer.dumps(login_history)

    response.set_cookie(
        LOGIN_HISTORY_COOKIE,
        cookie_value,
        max_age=SESSION_LIFETIME.total_seconds(),
        secure=current_app.config.get("SESSION_COOKIE_SECURE", False),
        httponly=True,
        samesite="Lax",
    )

    return response


def add_to_login_history(identifier):
    """Add a user identifier to the login history.

    The identifier is moved to the front if already present, ensuring
    the most recently used accounts appear first. The list is capped
    at MAX_LOGIN_HISTORY entries.

    :param identifier: The user identifier or username to remember
    """
    if not identifier:
        return

    # Use existing data from g if available (for tests), otherwise get from cookie
    login_history = getattr(g, "login_history_needs_update", None)
    if login_history is None:
        login_history = _get_login_history_from_cookie()

    if identifier in login_history:
        login_history.remove(identifier)

    login_history.insert(0, identifier)
    login_history = login_history[:MAX_LOGIN_HISTORY]

    session[LOGIN_HISTORY] = login_history
    g.login_history_needs_update = login_history


def get_login_history():
    """Retrieve the list of previously logged-in users from login history.

    Fetches user objects for stored identifiers, filtering out
    deleted or locked accounts. Uses get_user_from_login for
    consistent lookup logic.

    :return: List of User objects that are still valid
    """
    from canaille.core.auth import get_user_from_login

    identifiers = session.get(LOGIN_HISTORY)
    if identifiers is None:
        identifiers = _get_login_history_from_cookie()
        session[LOGIN_HISTORY] = identifiers

    users = []
    for identifier in identifiers:
        user = get_user_from_login(identifier)
        if user and (
            not current_app.backend.has_account_lockability() or not user.locked
        ):
            users.append(user)

    return users


def is_user_in_login_history(user_name):
    """Check if a user identifier is in the login history.

    :param user_name: The username/identifier to check
    :return: True if the user is in login history, False otherwise
    """
    identifiers = session.get(LOGIN_HISTORY)
    if identifiers is None:
        identifiers = _get_login_history_from_cookie()

    return user_name in identifiers


def remove_from_login_history(identifier):
    """Remove a user identifier from the login history.

    :param identifier: The username/identifier to remove
    """
    if not identifier:
        return

    # Use existing data from g if available (for tests), otherwise get from cookie
    login_history = getattr(g, "login_history_needs_update", None)
    if login_history is None:
        login_history = _get_login_history_from_cookie()

    if identifier in login_history:
        login_history.remove(identifier)

        if login_history:
            session[LOGIN_HISTORY] = login_history
        else:
            session.pop(LOGIN_HISTORY, None)

        g.login_history_needs_update = login_history



import re
import textwrap

from pydantic import BaseModel
from pydantic import ValidationError
from pydantic_core import PydanticUndefined
from pydantic_settings import BaseSettings

try:
    import tomlkit

    HAS_TOMLKIT = True
except ImportError:
    HAS_TOMLKIT = False


def sanitize_rst_text(text: str) -> str:
    """Remove inline RST syntax."""
    # Replace :foo:`~bar.Baz` with Baz
    text = re.sub(r":[\w:-]+:`~[\w\.]+\.(\w+)`", r"\1", text)

    # Replace :foo:`bar` and :foo:`bar <anything> with bar`
    text = re.sub(r":[\w:-]+:`([^`<]+)(?: <[^`>]+>)?`", r"\1", text)

    # Replace `label <URL>`_ with label (URL)
    text = re.sub(r"`([^`<]+) <([^`>]+)>`_", r"\1 (\2)", text)

    # Replace ``foo`` with `foo`
    text = re.sub(r"``([^`]+)``", r"\1", text)

    # Remove RST directives
    text = re.sub(r"\.\. [\w-]+::( \w+)?\n\n", "", text)

    return text


def sanitize_comments(text: str, line_length: int) -> str:
    """Remove RST syntax and wrap the docstring so it displays well as TOML comments."""

    def is_code_block(text: str) -> bool:
        return all(line.startswith("    ") for line in text.splitlines())

    def is_list(text: str) -> bool:
        return all(
            line.startswith("-") or line.startswith("  ") for line in text.splitlines()
        )

    text = sanitize_rst_text(text)
    paragraphs = text.split("\n\n")
    paragraphs = [
        textwrap.fill(paragraph, width=line_length)
        if not is_code_block(paragraph) and not is_list(paragraph)
        else paragraph
        for paragraph in paragraphs
    ]
    text = "\n\n".join(paragraphs)
    return text


def export_object_to_toml(
    obj,
    with_comments: bool = True,
    with_defaults: bool = True,
    line_length: int = 80,
):
    """Create a tomlkit document from an object."""

    def is_complex(obj) -> bool:
        return isinstance(obj, list | dict | BaseModel | BaseSettings)

    if isinstance(obj, BaseModel | BaseSettings):
        doc = tomlkit.document() if isinstance(obj, BaseSettings) else tomlkit.table()
        for field_name, field_info in obj.__class__.model_fields.items():
            field_value = getattr(obj, field_name)
            display_value = field_value is not None and (
                isinstance(field_value, BaseModel | BaseSettings)
                or field_value != field_info.default
            )
            display_comments = with_comments and field_info.description
            display_default_value = (
                with_defaults
                and not is_complex(field_info.default)
                and not is_complex(field_value)
            )

            if display_comments and (display_default_value or display_value):
                sanitized = sanitize_comments(field_info.description, line_length)
                for line in sanitized.splitlines():
                    doc.add(tomlkit.comment(line))

            if display_default_value:
                parsed = (
                    tomlkit.item(field_info.default).as_string()
                    if field_info.default is not None
                    and field_info.default is not PydanticUndefined
                    else ""
                )
                doc.add(tomlkit.comment(f"{field_name} = {parsed}".strip()))

            sub_value = export_object_to_toml(field_value)
            if display_value:
                doc.add(field_name, sub_value)
            doc.add(tomlkit.nl())
        return doc

    elif isinstance(obj, list):
        max_inline_items = 4
        is_multiline = len(obj) > max_inline_items or all(
            is_complex(item) for item in obj
        )
        doc = tomlkit.array().multiline(is_multiline)

        for item in obj:
            sub_value = export_object_to_toml(item)
            doc.append(sub_value)
        return doc

    elif isinstance(obj, dict):
        inline = all(not is_complex(item) for item in obj.values())
        doc = tomlkit.inline_table() if inline else tomlkit.table()
        for key, value in obj.items():
            sub_value = export_object_to_toml(value)
            doc.add(key, sub_value)
        return doc

    else:
        return obj


def export_config(model: BaseSettings):
    doc = export_object_to_toml(model)
    content = tomlkit.dumps(doc)

    # Remove end-of-line spaces
    content = re.sub(r" +\n", "\n", content)

    # Remove multiple new-lines
    content = re.sub(r"\n\n+", "\n\n", content)

    # Remove end-of-file new-line
    content = re.sub(r"\n+\Z", "\n", content)

    return content


def example_settings(model: type[BaseModel]) -> BaseModel | None:
    """Init a pydantic BaseModel with values passed as Field 'examples'."""
    data = {
        field_name: field_info.examples[0]
        for field_name, field_info in model.model_fields.items()
        if field_info.examples
    }
    try:
        return model.model_validate(data)
    except ValidationError:  # pragma: no cover
        return None



import datetime
from unittest import mock

import pytest
import time_machine
from flask import g

from canaille.app import models
from canaille.app.session import UserSession


def test_index(testclient, user, backend):
    res = testclient.get("/", status=302)
    assert res.location == "/login"

    g.session = UserSession(user=user)
    res = testclient.get("/", status=302)
    assert res.location == "/profile/user"

    testclient.app.config["CANAILLE"]["ACL"]["DEFAULT"]["PERMISSIONS"] = []
    backend.reload(g.session.user)
    res = testclient.get("/", status=302)
    assert res.location == "/about"


def test_user_deleted_in_session(testclient, backend):
    user = models.User(
        formatted_name="Jake Doe",
        family_name="Jake",
        user_name="jake",
        emails=["jake@doe.test"],
        password="correct horse battery staple",
    )
    backend.save(user)

    with testclient.session_transaction() as session:
        session["sessions"] = [
            UserSession(
                user=user,
                last_login_datetime=datetime.datetime.now(datetime.timezone.utc),
            ).serialize()
        ]

    testclient.get("/profile/jake", status=200)

    backend.delete(user)
    del g.session

    testclient.get("/profile/jake", status=404)
    with testclient.session_transaction() as session:
        assert not session.get("sessions")


def test_impersonate(testclient, logged_admin, user):
    res = testclient.get("/", status=302).follow(status=200).click("Account settings")
    assert "admin" == res.form["user_name"].value

    res = (
        testclient.get("/impersonate/user", status=302)
        .follow(status=302)
        .follow(status=200)
        .click("Account settings")
    )
    assert "user" == res.form["user_name"].value

    testclient.get("/logout", status=302).follow(status=302).follow(status=200)

    res = testclient.get("/", status=302).follow(status=200).click("Account settings")
    assert "admin" == res.form["user_name"].value


def test_admin_self_deletion(testclient, backend):
    admin = models.User(
        formatted_name="Temp admin",
        family_name="admin",
        user_name="temp",
        emails=["temp@temp.test"],
        password="admin",
    )
    backend.save(admin)
    with testclient.session_transaction() as sess:
        sess["sessions"] = [
            UserSession(
                user=admin,
                last_login_datetime=datetime.datetime.now(datetime.timezone.utc),
            ).serialize()
        ]

    res = testclient.get("/profile/temp/settings")
    res = res.form.submit(name="action", value="confirm-delete")
    res = (
        res.form.submit(name="action", value="delete", status=302)
        .follow(status=302)
        .follow(status=200)
    )

    assert backend.get(models.User, user_name="temp") is None

    with testclient.session_transaction() as sess:
        assert not sess.get("sessions")


def test_user_self_deletion(testclient, backend):
    user = models.User(
        formatted_name="Temp user",
        family_name="user",
        user_name="temp",
        emails=["temp@temp.test"],
        password="correct horse battery staple",
    )
    backend.save(user)
    with testclient.session_transaction() as sess:
        sess["sessions"] = [
            UserSession(
                user=user,
                last_login_datetime=datetime.datetime.now(datetime.timezone.utc),
            ).serialize()
        ]

    testclient.app.config["CANAILLE"]["ACL"]["DEFAULT"]["PERMISSIONS"] = ["edit_self"]
    res = testclient.get("/profile/temp/settings")
    res.mustcontain(no="Delete my account")

    testclient.app.config["CANAILLE"]["ACL"]["DEFAULT"]["PERMISSIONS"] = [
        "edit_self",
        "delete_account",
    ]
    # Simulate an app restart
    backend.reload(user)

    res = testclient.get("/profile/temp/settings")
    res.mustcontain("Delete my account")
    res = res.form.submit(name="action", value="confirm-delete")
    res = (
        res.form.submit(name="action", value="delete", status=302)
        .follow(status=302)
        .follow(status=200)
    )

    assert backend.get(models.User, user_name="temp") is None

    with testclient.session_transaction() as sess:
        assert not sess.get("sessions")

    testclient.app.config["CANAILLE"]["ACL"]["DEFAULT"]["PERMISSIONS"] = []


def test_account_locking(user, backend):
    assert not user.locked
    assert not user.lock_date
    assert backend.check_user_password(user, "correct horse battery staple") == (
        True,
        None,
    )

    user.lock_date = datetime.datetime.now(datetime.timezone.utc)
    assert user.locked
    backend.save(user)
    assert user.locked
    assert backend.get(models.User, id=user.id).locked
    assert backend.check_user_password(user, "correct horse battery staple") == (
        False,
        "Your account has been locked.",
    )

    user.lock_date = None
    backend.save(user)
    assert not user.locked
    assert not backend.get(models.User, id=user.id).locked
    assert backend.check_user_password(user, "correct horse battery staple") == (
        True,
        None,
    )


def test_account_locking_past_date(user, backend):
    assert not user.locked
    assert not user.lock_date
    assert backend.check_user_password(user, "correct horse battery staple") == (
        True,
        None,
    )

    user.lock_date = datetime.datetime.now(datetime.timezone.utc).replace(
        microsecond=0
    ) - datetime.timedelta(days=30)
    backend.save(user)
    assert user.locked
    assert backend.get(models.User, id=user.id).locked
    assert backend.check_user_password(user, "correct horse battery staple") == (
        False,
        "Your account has been locked.",
    )


def test_account_locking_future_date(user, backend):
    assert not user.locked
    assert not user.lock_date
    assert backend.check_user_password(user, "correct horse battery staple") == (
        True,
        None,
    )

    user.lock_date = datetime.datetime.now(datetime.timezone.utc).replace(
        microsecond=0
    ) + datetime.timedelta(days=365 * 4)
    backend.save(user)
    assert not user.locked
    assert not backend.get(models.User, id=user.id).locked
    assert backend.check_user_password(user, "correct horse battery staple") == (
        True,
        None,
    )


def test_account_locked_during_session(testclient, logged_user, backend):
    logged_user.lock_date = datetime.datetime.now(datetime.timezone.utc)
    backend.save(logged_user)
    testclient.get("/profile/user/settings", status=403)


def test_expired_password_redirection_and_register_new_password_for_memory_and_sql(
    testclient,
    logged_user,
    user,
    backend,
    admin,
):
    """time_machine does not work with ldap."""
    if "ldap" in backend.__class__.__module__:
        pytest.skip()

    testclient.app.config["WTF_CSRF_ENABLED"] = False
    backend.reload(logged_user)
    res = testclient.get("/profile/user/settings", status=200)
    res.form["password1"] = "123456789"
    res.form["password2"] = "123456789"

    with time_machine.travel("2020-01-01 01:00:00+00:00", tick=False) as traveller:
        res = res.form.submit(name="action", value="edit-settings")

        testclient.app.config["CANAILLE"]["PASSWORD_LIFETIME"] = "P5D"

        traveller.shift(datetime.timedelta(days=5, minutes=1))

        backend.reload(g.session.user)

        res = testclient.get("/profile/user/settings")

        testclient.get("/reset/admin", status=403)

        assert (
            "info",
            "Your password has expired, please choose a new password.",
        ) in res.flashes
        assert res.location == "/reset/user"

        backend.reload(logged_user)
        res = testclient.get("/reset/user")

        res.form["password"] = "foobarbaz"
        res.form["confirmation"] = "foobarbaz"
        res = res.form.submit()
        assert ("success", "Your password has been updated successfully") in res.flashes


@mock.patch("canaille.core.models.User.has_expired_password")
def test_expired_password_redirection_and_register_new_password_for_ldap_sql_and_memory(
    has_expired,
    testclient,
    logged_user,
    user,
    backend,
    admin,
):
    """time_machine does not work with ldap."""
    has_expired.return_value = False
    assert user.password_last_update is None
    res = testclient.get("/profile/user/settings", status=200)
    res.form["password1"] = "123456789"
    res.form["password2"] = "123456789"
    res = res.form.submit(name="action", value="edit-settings")
    backend.reload(logged_user)
    assert user.password_last_update is not None

    has_expired.return_value = True
    res = testclient.get("/profile/user/settings")
    testclient.get("/reset/admin", status=403)
    assert (
        "info",
        "Your password has expired, please choose a new password.",
    ) in res.flashes
    assert res.location == "/reset/user"
    backend.reload(logged_user)
    backend.reload(g.session.user)
    backend.reload(user)

    res = testclient.get("/reset/user")

    res.form["password"] = "foobarbaz"
    res.form["confirmation"] = "foobarbaz"
    res = res.form.submit()
    assert ("success", "Your password has been updated successfully") in res.flashes


def test_not_expired_password_or_wrong_user_redirection(
    testclient, logged_user, user, backend, admin
):
    assert user.password_last_update is None
    res = testclient.get("/profile/user/settings", status=200)
    res.form["password1"] = "123456789"
    res.form["password2"] = "123456789"
    res = res.form.submit(name="action", value="edit-settings")
    backend.reload(logged_user)
    assert user.password_last_update is not None

    def test_two_redirections(password_lifetime):
        testclient.app.config["CANAILLE"]["PASSWORD_LIFETIME"] = password_lifetime
        testclient.get("/reset/user", status=403)
        testclient.get("/reset/admin", status=403)

    test_two_redirections(None)

    testclient.app.config["CANAILLE"]["PASSWORD_LIFETIME"] = "PT0S"
    res = testclient.get("/profile/user/settings")
    assert (
        "info",
        "Your password has expired, please choose a new password.",
    ) in res.flashes
    assert res.location == "/reset/user"

    testclient.app.config["CANAILLE"]["PASSWORD_LIFETIME"] = "P1D"
    res = testclient.get("/profile/user/settings")
    res.form["password1"] = "123456789"
    res.form["password2"] = "123456789"
    res = res.form.submit(name="action", value="edit-settings")

    test_two_redirections("P1D")


def test_expired_password_needed_without_current_user(testclient, user):
    testclient.get("/reset/user", status=403)



"""Unique OIDC tokens.

Revision ID: 1755869767
Revises: 1745234836
Create Date: 2025-08-22 15:36:07.150344

"""

from collections.abc import Sequence

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "1755869767"
down_revision: str | None = "1745234836"
branch_labels: str | Sequence[str] | None = ()
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    with op.batch_alter_table("token") as batch_op:
        batch_op.create_unique_constraint("uq_token_refresh_token", ["refresh_token"])
        batch_op.create_unique_constraint("uq_token_access_token", ["access_token"])


def downgrade() -> None:
    with op.batch_alter_table("token") as batch_op:
        batch_op.drop_constraint("uq_token_refresh_token", type_="unique")
        batch_op.drop_constraint("uq_token_access_token", type_="unique")



from flask import Blueprint
from flask import abort
from flask import request

from canaille.app import models
from canaille.app.flask import render_htmx_template
from canaille.app.flask import user_needed
from canaille.app.forms import TableForm
from canaille.app.templating import render_template

bp = Blueprint("authorizations", __name__, url_prefix="/admin/authorization")


@bp.route("/", methods=["GET", "POST"])
@user_needed("manage_oidc")
def index(user):
    table_form = TableForm(models.AuthorizationCode, formdata=request.form)
    if request.form and request.form.get("page") and not table_form.validate():
        abort(404)

    return render_htmx_template(
        "oidc/authorization_list.html",
        menuitem="admin",
        table_form=table_form,
    )


@bp.route("/<authorizationcode:authorization>", methods=["GET", "POST"])
@user_needed("manage_oidc")
def view(user, authorization):
    return render_template(
        "oidc/authorization_view.html",
        authorization=authorization,
        menuitem="admin",
    )



from urllib.parse import parse_qs
from urllib.parse import urlsplit

from joserfc import jwt

from canaille.app import models
from canaille.oidc.oauth import setup_oauth

from . import client_credentials


def test_token_default_expiration_date(
    testclient, logged_user, client, server_jwk, backend
):
    res = testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="code",
            client_id=client.client_id,
            scope="openid profile email groups address phone",
            nonce="somenonce",
            redirect_uri="https://client.test/redirect1",
        ),
        status=200,
    )

    res = res.form.submit(name="answer", value="accept", status=302)
    params = parse_qs(urlsplit(res.location).query)
    code = params["code"][0]
    authcode = backend.get(models.AuthorizationCode, code=code)
    assert authcode.lifetime == 84400

    res = testclient.post(
        "/oauth/token",
        params=dict(
            grant_type="authorization_code",
            code=code,
            scope="openid profile email groups address phone",
            redirect_uri=client.redirect_uris[0],
        ),
        headers={"Authorization": f"Basic {client_credentials(client)}"},
        status=200,
    )

    assert res.json["expires_in"] == 864000

    access_token = res.json["access_token"]
    token = backend.get(models.Token, access_token=access_token)
    assert token.lifetime == 864000

    claims = jwt.decode(access_token, server_jwk)
    assert claims.claims["exp"] - claims.claims["iat"] == 864000

    id_token = res.json["id_token"]
    claims = jwt.decode(id_token, server_jwk)
    assert claims.claims["exp"] - claims.claims["iat"] == 3600

    consents = backend.query(models.Consent, client=client, subject=logged_user)
    for consent in consents:
        backend.delete(consent)


def test_token_custom_expiration_date(
    testclient, logged_user, client, server_jwk, backend
):
    testclient.app.config["OAUTH2_TOKEN_EXPIRES_IN"] = {
        "authorization_code": 1000,
        "implicit": 2000,
        "password": 3000,
        "client_credentials": 4000,
        "urn:ietf:params:oauth:grant-type:jwt-bearer": 5000,
    }
    setup_oauth(testclient.app)

    res = testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="code",
            client_id=client.client_id,
            scope="openid profile email groups address phone",
            nonce="somenonce",
            redirect_uri="https://client.test/redirect1",
        ),
        status=200,
    )

    res = res.form.submit(name="answer", value="accept", status=302)
    params = parse_qs(urlsplit(res.location).query)
    code = params["code"][0]
    authcode = backend.get(models.AuthorizationCode, code=code)
    assert authcode.lifetime == 84400

    res = testclient.post(
        "/oauth/token",
        params=dict(
            grant_type="authorization_code",
            code=code,
            scope="openid profile email groups address phone",
            redirect_uri=client.redirect_uris[0],
        ),
        headers={"Authorization": f"Basic {client_credentials(client)}"},
        status=200,
    )

    assert res.json["expires_in"] == 1000

    access_token = res.json["access_token"]
    token = backend.get(models.Token, access_token=access_token)
    assert token.lifetime == 1000

    claims = jwt.decode(access_token, server_jwk)
    assert claims.claims["exp"] - claims.claims["iat"] == 1000

    id_token = res.json["id_token"]
    claims = jwt.decode(id_token, server_jwk)
    lifetime = claims.claims["exp"] - claims.claims["iat"]
    assert lifetime == 3600

    consents = backend.query(models.Consent, client=client, subject=logged_user)
    for consent in consents:
        backend.delete(consent)



def test_group_permissions_by_dn(testclient, user, foo_group, backend):
    assert not user.can_manage_users

    testclient.app.config["CANAILLE"]["ACL"]["ADMIN"]["FILTER"] = {
        "groups": foo_group.dn
    }
    backend.reload(user)

    assert user.can_manage_users


def test_group_permissions_str(testclient, user, foo_group, backend):
    assert not user.can_manage_users

    testclient.app.config["CANAILLE"]["ACL"]["ADMIN"]["FILTER"] = (
        f"memberOf={foo_group.dn}"
    )
    backend.reload(user)

    assert user.can_manage_users



from canaille.app import validate_uri


def test_validate_uri():
    assert validate_uri("https://canaille.test")
    assert validate_uri("http://canaille.test")
    assert validate_uri("scheme.with.dots://canaille.tld")
    assert validate_uri("scheme.with.dots://localhost")
    assert validate_uri("scheme.with.dots://oauth")
    assert validate_uri("http://127.0.0.1")
    assert validate_uri("http://127.0.0.1:8000")
    assert not validate_uri("data://canaille.test")
    assert not validate_uri("file://canaille.test")
    assert not validate_uri("javascript:alert()")
    assert not validate_uri("invalid")



def setup_sentry(app):  # pragma: no cover
    if not app.config["CANAILLE"]["SENTRY_DSN"]:
        return None

    try:
        import sentry_sdk
        from sentry_sdk.integrations.flask import FlaskIntegration

    except Exception:
        return None

    sentry_sdk.init(
        dsn=app.config["CANAILLE"]["SENTRY_DSN"], integrations=[FlaskIntegration()]
    )
    return sentry_sdk



"""oidc client require_signed_request_object attribute.

Revision ID: 1745234836
Revises: 1743675164
Create Date: 2025-04-21 13:27:16.520455
"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "1745234836"
down_revision: str | None = "1743675164"
branch_labels: str | Sequence[str] | None = ()
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "client",
        sa.Column("require_signed_request_object", sa.Boolean(), nullable=True),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("client", "require_signed_request_object")
    # ### end Alembic commands ###



import datetime
import logging
from unittest import mock

import pytest


def test_password_forgotten_disabled(smtpd, testclient, user):
    testclient.app.config["CANAILLE"]["ENABLE_PASSWORD_RECOVERY"] = False

    testclient.get("/reset", status=404)
    testclient.get("/reset/user/hash", status=404)
    testclient.get("/reset-code/user", status=404)

    res = testclient.get("/login")
    res.mustcontain(no="Forgotten password")


def test_password_forgotten(smtpd, testclient, user, caplog):
    res = testclient.get("/reset", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=200)
    assert (
        "success",
        "A password reset link has been sent at your email address. You should receive "
        "it within a few minutes.",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Sending a reset password mail to john@doe.test for user",
    ) in caplog.record_tuples
    res.mustcontain("Send again")

    assert len(smtpd.messages) == 1


def test_password_forgotten_multiple_mails(smtpd, testclient, user, backend, caplog):
    user.emails = ["foo@bar.test", "foo@baz.test", "foo@foo.com"]
    backend.save(user)

    res = testclient.get("/reset", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=200)
    assert (
        "success",
        "A password reset link has been sent at your email address. You should receive "
        "it within a few minutes.",
    ) in res.flashes
    for email in user.emails:
        assert (
            "canaille",
            logging.SECURITY,
            f"Sending a reset password mail to {email} for user",
        ) in caplog.record_tuples
    res.mustcontain("Send again")

    assert len(smtpd.messages) == 3
    assert [message["X-RcptTo"] for message in smtpd.messages] == user.emails


def test_password_forgotten_invalid_form(smtpd, testclient, user):
    res = testclient.get("/reset", status=200)

    res.form["login"] = ""
    res = res.form.submit(status=200)
    assert ("error", "Could not send the password reset link.") in res.flashes

    assert len(smtpd.messages) == 0


def test_password_forgotten_invalid(smtpd, testclient, user):
    testclient.app.config["CANAILLE"]["HIDE_INVALID_LOGINS"] = True
    res = testclient.get("/reset", status=200)

    res.form["login"] = "i-dont-really-exist"
    res = res.form.submit(status=200)
    assert (
        "success",
        "A password reset link has been sent at your email address. "
        "You should receive it within a few minutes.",
    ) in res.flashes
    res.mustcontain(no="The login 'i-dont-really-exist' does not exist")

    testclient.app.config["CANAILLE"]["HIDE_INVALID_LOGINS"] = False
    res = testclient.get("/reset", status=200)

    res.form["login"] = "i-dont-really-exist"
    res = res.form.submit(status=200)
    assert (
        "success",
        "A password reset link has been sent at your email address. "
        "You should receive it within a few minutes.",
    ) not in res.flashes
    res.mustcontain("The login 'i-dont-really-exist' does not exist")

    assert len(smtpd.messages) == 0


def test_password_forgotten_invalid_when_user_cannot_self_edit(
    smtpd, testclient, user, backend
):
    testclient.app.config["CANAILLE"]["ACL"]["DEFAULT"]["PERMISSIONS"] = []
    backend.reload(user)

    testclient.app.config["CANAILLE"]["HIDE_INVALID_LOGINS"] = False
    res = testclient.get("/reset", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=200)
    assert (
        "success",
        "A password reset link has been sent at your email address. "
        "You should receive it within a few minutes.",
    ) not in res.flashes
    assert (
        "error",
        "The user 'John (johnny) Doe' does not have permissions to update their "
        "password. We cannot send a password reset email.",
    ) in res.flashes

    testclient.app.config["CANAILLE"]["HIDE_INVALID_LOGINS"] = True
    backend.reload(user)
    res = testclient.get("/reset", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=200)
    assert (
        "error",
        "The user 'John (johnny) Doe' does not have permissions to update their "
        "password. We cannot send a password reset email.",
    ) not in res.flashes
    assert (
        "success",
        "A password reset link has been sent at your email address. "
        "You should receive it within a few minutes.",
    ) in res.flashes

    assert len(smtpd.messages) == 0


@mock.patch("smtplib.SMTP")
def test_password_forgotten_mail_error(SMTP, smtpd, testclient, user):
    SMTP.side_effect = mock.Mock(side_effect=OSError("unit test mail error"))
    res = testclient.get("/reset", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=200, expect_errors=True)
    assert (
        "success",
        "A password reset link has been sent at your email address. "
        "You should receive it within a few minutes.",
    ) not in res.flashes
    assert (
        "error",
        "We encountered an issue while we sent the password recovery email.",
    ) in res.flashes
    res.mustcontain("Send again")

    assert len(smtpd.messages) == 0


def test_password_forgotten_user_disabled(smtpd, testclient, user, caplog, backend):
    """No mail should be sent if a disabled user asks for a password reset."""
    user.lock_date = datetime.datetime.now(datetime.timezone.utc)
    backend.save(user)

    res = testclient.get("/reset", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=200)
    assert (
        "success",
        "A password reset link has been sent at your email address. You should receive "
        "it within a few minutes.",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Sending a reset password mail to john@doe.test for user",
    ) not in caplog.record_tuples

    assert len(smtpd.messages) == 0


def test_password_forgotten_without_trusted_hosts(
    smtpd, testclient, user, caplog, backend
):
    testclient.app.config["TRUSTED_HOSTS"] = None
    res = testclient.get("/reset", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    assert (
        "success",
        "A password reset code has been sent at your email address. You should receive "
        "it within a few minutes.",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Sending a reset password mail to john@doe.test for user",
    ) in caplog.record_tuples

    assert len(smtpd.messages) == 1

    res = res.follow(status=200)

    backend.reload(user)
    code = user.one_time_password
    email_content = str(smtpd.messages[0].get_payload()[0]).replace("=\n", "")
    assert code in email_content

    main_form = res.forms[0]
    main_form["code"] = code
    res = main_form.submit(status=302)
    res.follow(status=200)

    assert res.location == f"/reset/{user.user_name}/{code}"


def test_password_forgotten_without_trusted_hosts_wrong_code(
    smtpd, testclient, user, caplog, backend
):
    testclient.app.config["TRUSTED_HOSTS"] = None
    res = testclient.get("/reset", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    main_form = res.forms[0]
    main_form["code"] = "111111"
    res = main_form.submit(status=200)

    assert ("error", "Invalid code.") in res.flashes


def test_password_forgotten_trying_to_access_code_page_with_trusted_hosts_enabled(
    smtpd, testclient, user
):
    testclient.get("/reset-code/user", status=404)


def test_password_forgotten_trying_to_access_code_page_when_user_cannot_self_edit(
    smtpd, testclient, user, backend
):
    testclient.app.config["CANAILLE"]["ACL"]["DEFAULT"]["PERMISSIONS"] = []
    testclient.app.config["TRUSTED_HOSTS"] = None
    backend.reload(user)

    testclient.app.config["CANAILLE"]["HIDE_INVALID_LOGINS"] = False
    res = testclient.get("/reset-code/user", status=302)
    assert (
        "error",
        "The user 'John (johnny) Doe' does not have permissions to update their "
        "password.",
    ) not in res.flashes


def test_password_reset_with_wrong_host(
    configuration, smtpd, testclient, backend, user
):
    res = testclient.get("/reset", status=200)
    res.form["login"] = "user"
    # Raises an attribute error but a Security Error is raised before during host validation, causing the attribute error later
    with pytest.raises(AttributeError):
        res.form.submit(headers={"Host": "test.test"}, status=400)



import datetime

from webtest import Upload

from canaille.app import models


def test_photo(testclient, user, jpeg_photo, backend):
    user.photo = jpeg_photo
    backend.save(user)
    backend.reload(user)

    res = testclient.get("/profile/user/photo")
    assert res.body == jpeg_photo
    assert res.last_modified == user.last_modified
    etag = res.etag
    assert etag

    res = testclient.get(
        "/profile/user/photo",
        headers={
            "If-Modified-Since": (
                res.last_modified + datetime.timedelta(days=1)
            ).strftime("%a, %d %b %Y %H:%M:%S UTC")
        },
        status=304,
    )
    assert not res.body

    res = testclient.get(
        "/profile/user/photo",
        headers={"If-None-Match": etag},
        status=304,
    )
    assert not res.body


def test_photo_invalid_user(testclient, user):
    testclient.get("/profile/invalid/photo", status=404)


def test_photo_absent(testclient, user):
    assert not user.photo
    testclient.get("/profile/user/photo", status=404)


def test_photo_invalid_path(testclient, user):
    testclient.get("/profile/user/invalid", status=404)


def test_photo_on_profile_edition(
    testclient,
    logged_user,
    jpeg_photo,
    backend,
):
    # Add a photo
    res = testclient.get("/profile/user", status=200)
    form = res.forms["baseform"]
    form["photo"] = Upload("logo.jpg", jpeg_photo)
    form["photo_delete"] = False
    res = form.submit(name="action", value="edit-profile")
    assert ("success", "Profile updated successfully.") in res.flashes
    res = res.follow()

    backend.reload(logged_user)

    assert logged_user.photo == jpeg_photo

    # No change
    res = testclient.get("/profile/user", status=200)
    form = res.forms["baseform"]
    form["photo_delete"] = False
    res = form.submit(name="action", value="edit-profile")
    assert ("success", "Profile updated successfully.") in res.flashes
    res = res.follow()

    backend.reload(logged_user)

    assert logged_user.photo == jpeg_photo

    # Photo deletion
    res = testclient.get("/profile/user", status=200)
    form = res.forms["baseform"]
    form["photo_delete"] = True
    res = form.submit(name="action", value="edit-profile")
    assert ("success", "Profile updated successfully.") in res.flashes
    res = res.follow()

    backend.reload(logged_user)

    assert logged_user.photo is None

    # Photo deletion AND upload, this should never happen
    res = testclient.get("/profile/user", status=200)
    form = res.forms["baseform"]
    form["photo"] = Upload("logo.jpg", jpeg_photo)
    form["photo_delete"] = True
    res = form.submit(name="action", value="edit-profile")
    assert ("success", "Profile updated successfully.") in res.flashes
    res = res.follow()

    backend.reload(logged_user)

    assert logged_user.photo is None


def test_photo_on_profile_creation(testclient, jpeg_photo, logged_admin, backend):
    res = testclient.get("/users", status=200)
    assert backend.get(models.User, user_name="foobar") is None
    res.mustcontain(no="foobar")

    res = testclient.get("/profile", status=200)
    form = res.forms["baseform"]
    form["photo"] = Upload("logo.jpg", jpeg_photo)
    form["user_name"] = "foobar"
    form["family_name"] = "Abitbol"
    form["emails-0"] = "george@abitbol.test"
    res = form.submit(name="action", value="edit-profile", status=302).follow(
        status=200
    )

    user = backend.get(models.User, user_name="foobar")
    assert user.photo == jpeg_photo
    backend.delete(user)


def test_photo_deleted_on_profile_creation(
    testclient, jpeg_photo, logged_admin, backend
):
    res = testclient.get("/users", status=200)
    assert backend.get(models.User, user_name="foobar") is None
    res.mustcontain(no="foobar")

    res = testclient.get("/profile", status=200)
    form = res.forms["baseform"]
    form["photo"] = Upload("logo.jpg", jpeg_photo)
    form["photo_delete"] = True
    form["user_name"] = "foobar"
    form["family_name"] = "Abitbol"
    form["emails-0"] = "george@abitbol.test"
    res = form.submit(name="action", value="edit-profile", status=302).follow(
        status=200
    )

    user = backend.get(models.User, user_name="foobar")
    assert user.photo is None
    backend.delete(user)



import importlib.metadata
import multiprocessing
import sys

import click
from flask.cli import FlaskGroup

import canaille.app.commands
import canaille.backends.commands
import canaille.core.commands
import canaille.oidc.commands
from canaille import create_app

version = importlib.metadata.version("canaille")


def create_cli_app():  # pragma: no cover
    # Force the non-application of migrations
    return create_app(init_backend=False)


@click.group(
    cls=FlaskGroup,
    create_app=create_cli_app,
    add_version_option=False,
    add_default_commands=False,
)
@click.version_option(version, prog_name="Canaille")
def cli():
    """Canaille management utilities."""


canaille.app.commands.register(cli)
canaille.backends.commands.register(cli)
canaille.core.commands.register(cli)
canaille.oidc.commands.register(cli)


if __name__ == "__main__":  # pragma: no cover
    # Needed by pyinstaller (not just on Windows)
    # https://pyinstaller.org/en/stable/common-issues-and-pitfalls.html#multi-processing
    if getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS"):
        multiprocessing.freeze_support()

    cli()



def install(config, debug=False):
    from joserfc import jwk

    if (
        not debug
        or config.get("CANAILLE_OIDC") is None
        or config["CANAILLE_OIDC"].get("ACTIVE_JWKS")
    ):
        return

    key = jwk.generate_key("RSA", 4096)
    config["CANAILLE_OIDC"]["ACTIVE_JWKS"] = [key.as_dict()]



"""Missing oidc client attributes.

Revision ID: 1742298790
Revises: 1742033764
Create Date: 2025-03-18 12:53:10.926898

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "1742298790"
down_revision: str | None = "1742033764"
branch_labels: str | Sequence[str] | None = ()
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "client",
        sa.Column("token_endpoint_auth_signing_alg", sa.String(), nullable=True),
    )
    op.add_column(
        "client", sa.Column("sector_identifier_uri", sa.String(), nullable=True)
    )
    op.add_column("client", sa.Column("subject_type", sa.String(), nullable=True))
    op.add_column("client", sa.Column("application_type", sa.String(), nullable=True))
    op.add_column(
        "client", sa.Column("id_token_signed_response_alg", sa.String(), nullable=True)
    )
    op.add_column(
        "client",
        sa.Column("id_token_encrypted_response_alg", sa.String(), nullable=True),
    )
    op.add_column(
        "client",
        sa.Column("id_token_encrypted_response_enc", sa.String(), nullable=True),
    )
    op.add_column(
        "client", sa.Column("userinfo_signed_response_alg", sa.String(), nullable=True)
    )
    op.add_column(
        "client",
        sa.Column("userinfo_encrypted_response_alg", sa.String(), nullable=True),
    )
    op.add_column(
        "client",
        sa.Column("userinfo_encrypted_response_enc", sa.String(), nullable=True),
    )
    op.add_column("client", sa.Column("default_max_age", sa.Integer(), nullable=True))
    op.add_column("client", sa.Column("require_auth_time", sa.Boolean(), nullable=True))
    op.add_column("client", sa.Column("default_acr_values", sa.JSON(), nullable=True))
    op.add_column("client", sa.Column("initiate_login_uri", sa.String(), nullable=True))
    op.add_column(
        "client", sa.Column("request_object_signing_alg", sa.String(), nullable=True)
    )
    op.add_column(
        "client", sa.Column("request_object_encryption_alg", sa.String(), nullable=True)
    )
    op.add_column(
        "client", sa.Column("request_object_encryption_enc", sa.String(), nullable=True)
    )
    op.add_column("client", sa.Column("request_uris", sa.JSON(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("client", "request_uris")
    op.drop_column("client", "request_object_encryption_enc")
    op.drop_column("client", "request_object_encryption_alg")
    op.drop_column("client", "request_object_signing_alg")
    op.drop_column("client", "initiate_login_uri")
    op.drop_column("client", "default_acr_values")
    op.drop_column("client", "require_auth_time")
    op.drop_column("client", "default_max_age")
    op.drop_column("client", "userinfo_encrypted_response_enc")
    op.drop_column("client", "userinfo_encrypted_response_alg")
    op.drop_column("client", "userinfo_signed_response_alg")
    op.drop_column("client", "id_token_encrypted_response_enc")
    op.drop_column("client", "id_token_encrypted_response_alg")
    op.drop_column("client", "id_token_signed_response_alg")
    op.drop_column("client", "application_type")
    op.drop_column("client", "subject_type")
    op.drop_column("client", "sector_identifier_uri")
    op.drop_column("client", "token_endpoint_auth_signing_alg")
    # ### end Alembic commands ###



"""Tests for TRUSTED_DOMAINS functionality."""

from canaille.app import models


def test_trusted_domains_property(testclient, backend):
    """Test that the trusted property works with TRUSTED_DOMAINS configuration."""
    client_localhost = models.Client(
        client_id="localhost-client",
        client_name="Localhost Client",
        client_uri="http://localhost:3000",
    )
    backend.save(client_localhost)

    client_trusted = models.Client(
        client_id="trusted-client",
        client_name="Trusted Client",
        client_uri="https://client.trusted.test",
    )
    backend.save(client_trusted)

    client_subdomain = models.Client(
        client_id="subdomain-client",
        client_name="Subdomain Client",
        client_uri="https://app.trusted.test",
    )
    backend.save(client_subdomain)

    client_untrusted = models.Client(
        client_id="untrusted-client",
        client_name="Untrusted Client",
        client_uri="https://evil.example.com",
    )
    backend.save(client_untrusted)

    client_no_uri = models.Client(
        client_id="no-uri-client",
        client_name="No URI Client",
    )
    backend.save(client_no_uri)

    assert client_localhost.trusted
    assert client_trusted.trusted
    assert client_subdomain.trusted
    assert not client_untrusted.trusted
    assert not client_no_uri.trusted


def test_trusted_domains_consent_bypass(testclient, logged_user, backend):
    """Test that trusted clients bypass the consent page."""
    client = models.Client(
        client_id="test-client",
        client_name="Test Client",
        client_uri="https://localhost:3000",
        redirect_uris=["https://localhost:3000/callback"],
        grant_types=["authorization_code"],
        response_types=["code"],
        scope=["openid", "profile"],
    )
    backend.save(client)
    client.audience = [client]
    backend.save(client)

    res = testclient.get(
        "/oauth/authorize",
        {
            "response_type": "code",
            "client_id": client.client_id,
            "scope": "openid profile",
            "redirect_uri": "https://localhost:3000/callback",
            "nonce": "test-nonce",
        },
        status=302,
    )

    assert res.location.startswith("https://localhost:3000/callback?code=")


def test_untrusted_domains_show_consent(testclient, logged_user, backend):
    """Test that untrusted clients show the consent page."""
    client = models.Client(
        client_id="untrusted-consent-client",
        client_name="Untrusted Client",
        client_uri="https://untrusted.example.com",
        redirect_uris=["https://untrusted.example.com/callback"],
        grant_types=["authorization_code"],
        response_types=["code"],
        scope=["openid", "profile"],
    )
    backend.save(client)
    client.audience = [client]
    backend.save(client)

    res = testclient.get(
        "/oauth/authorize",
        {
            "response_type": "code",
            "client_id": client.client_id,
            "scope": "openid profile",
            "redirect_uri": "https://untrusted.example.com/callback",
            "nonce": "untrusted-test-nonce",
        },
        status=200,
    )

    res.mustcontain("is requesting access to")
    res.mustcontain("Untrusted Client")


def test_wildcard_and_exact_domain_matching(testclient, backend):
    """Test wildcard domains and exact domain matching."""
    testclient.app.config["CANAILLE_OIDC"]["TRUSTED_DOMAINS"] = [
        ".dev.test",
        "exact.test",
    ]

    client_wildcard_sub = models.Client(
        client_id="wildcard-sub-client",
        client_name="Wildcard Sub Client",
        client_uri="https://app.dev.test",
    )
    backend.save(client_wildcard_sub)

    client_wildcard_exact = models.Client(
        client_id="wildcard-exact-client",
        client_name="Wildcard Exact Client",
        client_uri="https://dev.test",
    )
    backend.save(client_wildcard_exact)

    client_exact_sub = models.Client(
        client_id="exact-sub-client",
        client_name="Exact Sub Client",
        client_uri="https://sub.exact.test",
    )
    backend.save(client_exact_sub)

    client_exact_exact = models.Client(
        client_id="exact-exact-client",
        client_name="Exact Exact Client",
        client_uri="https://exact.test",
    )
    backend.save(client_exact_exact)

    client_no_match = models.Client(
        client_id="no-match-client",
        client_name="No Match Client",
        client_uri="https://app.untrusted.test",
    )
    backend.save(client_no_match)

    assert client_wildcard_sub.trusted
    assert client_wildcard_exact.trusted
    assert not client_exact_sub.trusted
    assert client_exact_exact.trusted
    assert not client_no_match.trusted


def test_empty_trusted_domains(testclient, backend):
    """Test that clients are not trusted when TRUSTED_DOMAINS is empty."""
    testclient.app.config["CANAILLE_OIDC"]["TRUSTED_DOMAINS"] = []

    client = models.Client(
        client_id="empty-domains-client",
        client_name="Test Client",
        client_uri="https://example.com",
    )
    backend.save(client)

    assert not client.trusted


def test_invalid_client_uri_hostname(testclient, backend):
    """Test that clients with invalid URI hostnames are not trusted."""
    testclient.app.config["CANAILLE_OIDC"]["TRUSTED_DOMAINS"] = ["example.com"]

    client = models.Client(
        client_id="invalid-uri-client",
        client_name="Test Client",
        client_uri="invalid-uri-without-hostname",
    )
    backend.save(client)

    assert not client.trusted



import datetime
import json
import os
import uuid

import pytest
from authlib.oidc.core import UserInfo
from authlib.oidc.core.grants.util import generate_id_token
from joserfc import jwk
from werkzeug.security import gen_salt

from canaille.app import models
from canaille.oidc.oauth import generate_user_claims
from canaille.oidc.oauth import get_jwt_config


@pytest.fixture
# For some reason all the params from the overridden fixture must be present here
# https://github.com/pytest-dev/pytest/issues/11075
def app(app, configuration, backend):
    os.environ["AUTHLIB_INSECURE_TRANSPORT"] = "true"
    yield app


@pytest.fixture(scope="session")
def server_jwk():
    key = jwk.generate_key("RSA", 1024)
    key.ensure_kid()
    return key


@pytest.fixture(scope="session")
def old_server_jwk():
    key = jwk.generate_key("RSA", 1024)
    key.ensure_kid()
    return key


@pytest.fixture
def configuration(configuration, server_jwk, old_server_jwk):
    configuration["CANAILLE_OIDC"] = {
        "ACTIVE_JWKS": [server_jwk.as_dict()],
        "INACTIVE_JWKS": [old_server_jwk.as_dict()],
        "TRUSTED_DOMAINS": [".localhost", "127.0.0.1", ".trusted.test"],
    }
    configuration["CANAILLE"]["LOGGING"]["loggers"]["authlib"] = {
        "level": "DEBUG",
        "handlers": ["default"],
    }
    return configuration


@pytest.fixture
def client_jwk():
    return jwk.generate_key("RSA", 1024)


@pytest.fixture
def client(testclient, trusted_client, backend, client_jwk):
    key_set = jwk.KeySet([client_jwk]).as_dict(private=False)
    c = models.Client(
        client_id=gen_salt(24),
        client_name="Some client",
        contacts=["contact@mydomain.test"],
        client_uri="https://client.test",
        redirect_uris=[
            "https://client.test/redirect1",
            "https://client.test/redirect2",
        ],
        logo_uri="https://client.test/logo.webp",
        client_id_issued_at=datetime.datetime.now(datetime.timezone.utc),
        client_secret=gen_salt(48),
        grant_types=[
            "password",
            "authorization_code",
            "implicit",
            "hybrid",
            "refresh_token",
            "client_credentials",
            "urn:ietf:params:oauth:grant-type:jwt-bearer",
        ],
        response_types=["code", "token", "id_token"],
        scope=["openid", "email", "profile", "groups", "address", "phone"],
        tos_uri="https://client.test/tos",
        policy_uri="https://client.test/policy",
        jwks=json.dumps(key_set),
        token_endpoint_auth_method="client_secret_basic",
        post_logout_redirect_uris=["https://client.test/disconnected"],
    )
    backend.save(c)
    c.audience = [c, trusted_client]
    backend.save(c)

    yield c
    backend.delete(c)


@pytest.fixture
def trusted_client(testclient, backend, client_jwk):
    key_set = jwk.KeySet([client_jwk]).as_dict(private=False)
    c = models.Client(
        client_id=gen_salt(24),
        client_name="Some other client",
        contacts=["contact@trusted.test"],
        client_uri="https://client.trusted.test",
        redirect_uris=[
            "https://client.trusted.test/redirect1",
            "https://client.trusted.test/redirect2",
        ],
        logo_uri="https://client.trusted.test/logo.webp",
        client_id_issued_at=datetime.datetime.now(datetime.timezone.utc),
        client_secret=gen_salt(48),
        grant_types=[
            "password",
            "authorization_code",
            "implicit",
            "hybrid",
            "refresh_token",
            "client_credentials",
            "urn:ietf:params:oauth:grant-type:jwt-bearer",
        ],
        response_types=["code", "token", "id_token"],
        scope=["openid", "profile", "groups"],
        tos_uri="https://client.trusted.test/tos",
        policy_uri="https://client.trusted.test/policy",
        jwks=json.dumps(key_set),
        token_endpoint_auth_method="client_secret_basic",
        post_logout_redirect_uris=["https://client.trusted.test/disconnected"],
    )
    backend.save(c)
    c.audience = [c]
    backend.save(c)

    yield c
    backend.delete(c)


@pytest.fixture
def authorization(testclient, user, client, backend):
    a = models.AuthorizationCode(
        authorization_code_id=gen_salt(48),
        code="my-code",
        client=client,
        subject=user,
        redirect_uri="https://client.test/redirect1",
        response_type="code",
        scope=["openid", "profile"],
        nonce="nonce",
        issue_date=datetime.datetime(2020, 1, 1, tzinfo=datetime.timezone.utc),
        lifetime=3600,
        challenge=gen_salt(48),
        challenge_method="plain",
    )
    backend.save(a)
    yield a
    backend.delete(a)


@pytest.fixture
def token(testclient, client, user, backend):
    t = models.Token(
        token_id=gen_salt(48),
        access_token=gen_salt(48),
        audience=[client],
        client=client,
        subject=user,
        refresh_token=gen_salt(48),
        scope=["openid", "profile"],
        issue_date=datetime.datetime.now(datetime.timezone.utc),
        lifetime=3600,
    )
    backend.save(t)
    yield t
    backend.delete(t)


@pytest.fixture
def oidc_token(testclient, client, backend):
    t = models.Token(
        token_id=gen_salt(48),
        access_token=gen_salt(48),
        audience=[client],
        client=client,
        refresh_token=gen_salt(48),
        scope=["openid", "profile"],
        issue_date=datetime.datetime.now(datetime.timezone.utc),
        lifetime=3600,
    )
    backend.save(t)
    yield t
    backend.delete(t)


@pytest.fixture
def id_token(testclient, client, user, backend):
    return generate_id_token(
        {},
        UserInfo(generate_user_claims(user)).filter(client.scope),
        aud=client.client_id,
        **get_jwt_config(None),
    )


@pytest.fixture
def consent(testclient, client, user, backend):
    t = models.Consent(
        consent_id=str(uuid.uuid4()),
        client=client,
        subject=user,
        scope=["openid", "profile"],
        issue_date=datetime.datetime.now(datetime.timezone.utc),
    )
    backend.save(t)
    yield t
    backend.delete(t)



import pytest
from flask_babel import refresh
from werkzeug.datastructures import LanguageAccept

from canaille.app.i18n import smart_language_match


def test_preferred_language(testclient, logged_user, backend):
    logged_user.preferred_language = None
    backend.save(logged_user)

    res = testclient.get("/profile/user", status=200)
    form = res.forms["baseform"]
    assert form["preferred_language"].value == "auto"
    assert res.pyquery("html")[0].attrib["lang"] == "en"
    res.mustcontain("My profile")
    res.mustcontain(no="Mon profil")

    form["preferred_language"] = "fr"
    res = form.submit(name="action", value="edit-profile")
    assert res.flashes == [("success", "Le profil a été mis à jour avec succès.")]
    res = res.follow()
    form = res.forms["baseform"]
    backend.reload(logged_user)
    assert logged_user.preferred_language == "fr"
    assert form["preferred_language"].value == "fr"
    assert res.pyquery("html")[0].attrib["lang"] == "fr"
    res.mustcontain(no="My profile")
    res.mustcontain("Mon profil")

    form["preferred_language"] = "en"
    res = form.submit(name="action", value="edit-profile")
    assert res.flashes == [("success", "Profile updated successfully.")]
    res = res.follow()
    form = res.forms["baseform"]
    backend.reload(logged_user)
    assert logged_user.preferred_language == "en"
    assert form["preferred_language"].value == "en"
    assert res.pyquery("html")[0].attrib["lang"] == "en"
    res.mustcontain("My profile")
    res.mustcontain(no="Mon profil")

    form["preferred_language"] = "auto"
    res = form.submit(name="action", value="edit-profile")
    assert res.flashes == [("success", "Profile updated successfully.")]
    res = res.follow()
    form = res.forms["baseform"]
    backend.reload(logged_user)
    assert logged_user.preferred_language is None
    assert form["preferred_language"].value == "auto"
    assert res.pyquery("html")[0].attrib["lang"] == "en"
    res.mustcontain("My profile")
    res.mustcontain(no="Mon profil")


def test_form_translations(testclient, logged_user, backend):
    logged_user.preferred_language = "fr"
    backend.save(logged_user)

    res = testclient.get("/profile/user", status=200)
    form = res.forms["baseform"]
    form["phone_numbers-0"] = "invalid"
    res = form.submit(name="action", value="edit-profile")

    res.mustcontain(no="Not a valid phone number")
    res.mustcontain("N’est pas un numéro de téléphone valid")


def test_language_config(testclient, logged_user, backend):
    logged_user.preferred_language = None
    backend.save(logged_user)

    res = testclient.get("/profile/user", status=200)
    assert res.pyquery("html")[0].attrib["lang"] == "en"
    res.mustcontain("My profile")
    res.mustcontain(no="Mon profil")

    testclient.app.config["CANAILLE"]["LANGUAGE"] = "fr"
    refresh()
    res = testclient.get("/profile/user", status=200)
    assert res.pyquery("html")[0].attrib["lang"] == "fr"
    res.mustcontain(no="My profile")
    res.mustcontain("Mon profil")


@pytest.mark.parametrize(
    "accept_list,available,expected",
    [
        # Regional variant fallback to base language
        (
            [("fr-FR", 1), ("en-US", 0.7), ("en", 0.3)],
            ["gl", "de", "fr", "es", "nb_NO", "en"],
            "fr",
        ),
        # Exact match with underscore variant
        (
            [("nb-NO", 1), ("en", 0.5)],
            ["gl", "de", "fr", "es", "nb_NO", "en"],
            "nb_NO",
        ),
        # Base language matches underscore variant
        (
            [("nb", 1), ("en", 0.5)],
            ["gl", "de", "fr", "es", "nb_NO", "en"],
            "nb_NO",
        ),
        # Exact match (non-English)
        (
            [("fr", 1), ("en", 0.5)],
            ["gl", "de", "fr", "es", "nb_NO", "en"],
            "fr",
        ),
        # English preference (exact match)
        (
            [("en-US", 1), ("en", 0.9), ("fr", 0.3)],
            ["gl", "de", "fr", "es", "nb_NO", "en"],
            "en",
        ),
        # No match, fallback to English
        (
            [("zh-CN", 1), ("ja-JP", 0.9)],
            ["gl", "de", "fr", "es", "nb_NO", "en"],
            "en",
        ),
    ],
)
def test_smart_language_match(accept_list, available, expected):
    accept = LanguageAccept(accept_list)
    result = smart_language_match(accept, available)
    assert result == expected



import datetime

from blinker import signal
from flask import current_app
from httpx import Client as httpx_client
from scim2_client import SCIMClientError
from scim2_client.engines.httpx import SyncSCIMClient
from scim2_models import SearchRequest
from werkzeug.security import gen_salt

from canaille.app import models
from canaille.backends import Backend
from canaille.scim.casting import group_from_canaille_to_scim
from canaille.scim.casting import user_from_canaille_to_scim


def user_from_canaille_to_scim_client(user, user_class, enterprise_user_class):
    scim_user = user_from_canaille_to_scim(user, user_class, enterprise_user_class)
    scim_user.external_id = user.id
    return scim_user


def group_from_canaille_to_scim_client(group, group_class, scim_client):
    scim_group = group_from_canaille_to_scim(group, group_class)
    scim_group.external_id = group.id

    distant_members = []
    User = scim_client.get_resource_model("User")
    for member in group.members:
        req = SearchRequest(filter=f'externalId eq "{member.id}"')
        response = scim_client.query(User, search_request=req)
        if response.resources:
            distant_members.append(response.resources[0])
        else:
            current_app.logger.warning(
                f"Unable to find user {member.user_name} from group {group.display_name} via SCIM"
            )

    scim_group.members = [
        group_class.Members(
            value=user.id,
            type="User",
            display=user.display_name,
            ref=user.meta.location,
        )
        for user in distant_members or []
    ] or None

    return scim_group


def get_or_create_token(client):
    """Retrieve or initialize a client token."""
    scim_tokens = Backend.instance.query(models.Token, client=client, subject=None)
    valid_scim_tokens = [
        token
        for token in scim_tokens
        if not token.is_expired() and not token.is_revoked()
    ]
    if valid_scim_tokens:
        scim_token = valid_scim_tokens[0]
    else:
        scim_token = models.Token(
            token_id=gen_salt(48),
            access_token=gen_salt(48),
            subject=None,
            audience=[client],
            client=client,
            refresh_token=gen_salt(48),
            scope=["openid", "profile"],
            issue_date=datetime.datetime.now(datetime.timezone.utc),
            lifetime=3600,
        )
        Backend.instance.save(scim_token)
    return scim_token


def initiate_scim_client(client):
    """Set up the scim2-client object that will perform the SCIM2 requests."""
    if not client:  # pragma: no cover
        return None

    access_token = get_or_create_token(client).access_token
    client_httpx = httpx_client(
        base_url=client.client_uri,
        headers={"Authorization": f"Bearer {access_token}"},
    )
    scim = SyncSCIMClient(client_httpx)
    try:
        scim.discover()
    except SCIMClientError:
        current_app.logger.info(
            f"SCIM protocol not supported by client {client.client_name}"
        )
        return None
    return scim


def execute_scim_user_action(scim, user, client_name, method):
    """Create/update/delete a distant user with SCIM requests."""
    User = scim.get_resource_model("User")

    req = SearchRequest(filter=f'externalId eq "{user.id}"')
    response = scim.query(User, search_request=req)
    distant_scim_user = response.resources[0] if response.resources else None

    if method == "delete" and distant_scim_user:
        try:
            scim.delete(User, distant_scim_user.id)
        except:
            current_app.logger.warning(
                f"SCIM User {user.user_name} delete for client {client_name} failed"
            )
    elif method == "save":
        EnterpriseUser = User.get_extension_model("EnterpriseUser")
        scim_user = user_from_canaille_to_scim_client(user, User, EnterpriseUser)
        if not distant_scim_user:
            try:
                scim.create(scim_user)
            except Exception:
                current_app.logger.warning(
                    f"SCIM User {user.user_name} creation for client {client_name} failed"
                )
        else:
            scim_user.id = distant_scim_user.id
            try:
                scim.replace(scim_user)
            except Exception:
                current_app.logger.warning(
                    f"SCIM User {user.user_name} update for client {client_name} failed"
                )


def propagate_user_scim_modification(user, method):
    """After a user edition/deletion, broadcast the event to all the clients."""
    for client in get_clients_to_notify(user):
        scim = initiate_scim_client(client)
        if scim:
            execute_scim_user_action(scim, user, client.client_name, method)


def execute_scim_group_action(scim, group, client_name, method):
    """Create/update/delete a distant group with SCIM requests."""
    Group = scim.get_resource_model("Group")

    req = SearchRequest(filter=f'externalId eq "{group.id}"')
    response = scim.query(Group, search_request=req)
    scim_group = response.resources[0] if response.resources else None

    if method == "delete" and scim_group:
        try:
            scim.delete(Group, scim_group.id)
        except Exception:
            current_app.logger.warning(
                f"SCIM Group {group.display_name} delete for client {client_name} failed"
            )
    elif method == "save":  # pragma: no branch
        group = group_from_canaille_to_scim_client(group, Group, scim)
        if not scim_group:
            try:
                scim.create(group)
            except Exception:
                current_app.logger.warning(
                    f"SCIM Group {group.display_name} creation for client {client_name} failed"
                )
        else:
            group.id = scim_group.id
            try:
                scim.replace(group)
            except Exception:
                current_app.logger.warning(
                    f"SCIM Group {group.display_name} update for client {client_name} failed"
                )


def propagate_group_scim_modification(group, method):
    """After a group edition/deletion, broadcast the event to all the clients."""
    notifiable_clients = set()
    for member in group.members:
        notifiable_clients.update(get_clients_to_notify(member))

    for client in notifiable_clients:
        scim = initiate_scim_client(client)
        if scim:
            execute_scim_group_action(scim, group, client.client_name, method)


def get_clients_to_notify(user):
    """Return a list of clients that should be notified of updates on 'user'."""
    consents = Backend.instance.query(models.Consent, subject=user)
    consented_clients = {t.client for t in consents}
    trusted_clients = [
        client
        for client in Backend.instance.query(models.Client)
        if client.trusted and client not in consented_clients
    ]
    return list(consented_clients) + list(trusted_clients)


def after_user_query(user):
    user.old_groups = user.groups.copy()


def after_user_save(user, data):
    """Update the user object on the distant applications.

    As the user 'groups' attribute is read-only, it is needed
    to manually update the distant groups when the user group membership has changed.
    """
    propagate_user_scim_modification(user, method="save")

    old_groups = getattr(user, "old_groups", [])
    for group in set(old_groups) ^ set(user.groups):
        Backend.instance.reload(group)
        propagate_group_scim_modification(group, "save")

    user.old_groups = user.groups.copy()


def before_user_delete(user, data):
    propagate_user_scim_modification(user, method="delete")


def after_user_delete(user, data):
    for group in user.groups:
        Backend.instance.reload(group)
        propagate_group_scim_modification(group, "save")


def after_group_save(group, data):
    propagate_group_scim_modification(group, method="save")


def before_group_delete(group, data):
    propagate_group_scim_modification(group, method="delete")


def setup_scim_client():
    signal("after_user_query").connect(after_user_query)
    signal("after_user_save").connect(after_user_save)
    signal("before_user_delete").connect(before_user_delete)
    signal("after_user_delete").connect(after_user_delete)
    signal("after_group_save").connect(after_group_save)
    signal("before_group_delete").connect(before_group_delete)



import json

from joserfc import jwk

from canaille.commands import cli


def test_create_rsa_key(cli_runner):
    """Test the default RSA key creation command."""
    res = cli_runner.invoke(cli, ["jwk", "create", "rsa"], catch_exceptions=False)
    assert res.exit_code == 0, res.stdout
    payload = json.loads(res.stdout)
    jwk.import_key(payload, "RSA")


def test_create_rsa_key_with_size(cli_runner):
    """Test the RSA key creation command with a size parameter."""
    res = cli_runner.invoke(
        cli, ["jwk", "create", "rsa", "--size", "1024"], catch_exceptions=False
    )
    assert res.exit_code == 0, res.stdout
    payload = json.loads(res.stdout)
    jwk.import_key(payload, "RSA")


def test_create_oct_key(cli_runner):
    """Test the default Oct key creation command."""
    res = cli_runner.invoke(cli, ["jwk", "create", "oct"], catch_exceptions=False)
    assert res.exit_code == 0, res.stdout
    payload = json.loads(res.stdout)
    jwk.import_key(payload, "oct")


def test_create_oct_key_with_size(cli_runner):
    """Test the Oct key creation command with a size parameter."""
    res = cli_runner.invoke(
        cli, ["jwk", "create", "oct", "--size", "128"], catch_exceptions=False
    )
    assert res.exit_code == 0, res.stdout
    payload = json.loads(res.stdout)
    jwk.import_key(payload, "oct")


def test_create_ec_key(cli_runner):
    """Test the default EC key creation command."""
    res = cli_runner.invoke(cli, ["jwk", "create", "ec"], catch_exceptions=False)
    assert res.exit_code == 0, res.stdout
    payload = json.loads(res.stdout)
    jwk.import_key(payload, "EC")


def test_create_ec_key_with_crv(cli_runner):
    """Test the EC key creation command with a crv param."""
    res = cli_runner.invoke(
        cli, ["jwk", "create", "ec", "--crv", "secp256k1"], catch_exceptions=False
    )
    assert res.exit_code == 0, res.stdout
    payload = json.loads(res.stdout)
    jwk.import_key(payload, "EC")


def test_create_opk_key(cli_runner):
    """Test the default OPK key creation command."""
    res = cli_runner.invoke(cli, ["jwk", "create", "okp"], catch_exceptions=False)
    assert res.exit_code == 0, res.stdout
    payload = json.loads(res.stdout)
    jwk.import_key(payload, "OKP")


def test_create_opk_key_with_crv(cli_runner):
    """Test the OPK key creation command with a crv param."""
    res = cli_runner.invoke(
        cli, ["jwk", "create", "okp", "--crv", "X448"], catch_exceptions=False
    )
    assert res.exit_code == 0, res.stdout
    payload = json.loads(res.stdout)
    jwk.import_key(payload, "OKP")



import logging
from unittest import mock

from canaille.core.auth import AuthenticationSession


def test_no_auth_session_not_logged_in(testclient, user):
    """Non-logged users should not be able to access the password auth form if an authentication session has not been properly started."""
    res = testclient.get("/auth/password", status=302)
    assert res.location == "/login"
    assert res.flashes == [
        ("warning", "Cannot remember the login you attempted to sign in with.")
    ]


def test_no_auth_session_logged_in(testclient, logged_user):
    """Logged users should not be able to access the password auth form if an authentication session has not been properly started."""
    res = testclient.get("/auth/password", status=302)
    assert res.location == "/"


def test_not_password_step(testclient, user):
    """Users reaching the password form while this is not the right auth step in their flow should be redirected there."""
    with testclient.session_transaction() as session:
        session["auth"] = AuthenticationSession(
            user_name="user",
            remaining=["email", "password"],
        ).serialize()

    res = testclient.get("/auth/password", status=302)
    assert res.location == "/auth/email"


def test_signin_and_out(testclient, user, caplog):
    with testclient.session_transaction() as session:
        assert not session.get("sessions")

    res = testclient.get("/login", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    with testclient.session_transaction() as session:
        assert "user" == session["auth"]["user_name"]

    res.form["password"] = "correct horse battery staple"
    res = res.form.submit(status=302)

    assert (
        "success",
        "Connection successful. Welcome Johnny",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Successful authentication for user",
    ) in caplog.record_tuples
    res = res.follow(status=302)
    res = res.follow(status=200)

    with testclient.session_transaction() as session:
        assert [{"user": user.id, "last_login_datetime": mock.ANY}] == session.get(
            "sessions"
        )
        assert "auth" not in session

    res = testclient.get("/login", status=302)

    res = testclient.get("/logout")
    assert (
        "success",
        "You have been disconnected. See you next time Johnny",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Logout user",
    ) in caplog.record_tuples
    res = res.follow(status=302)
    res = res.follow(status=200)


def test_signin_wrong_password(testclient, user, caplog):
    with testclient.session_transaction() as session:
        assert not session.get("sessions")

    res = testclient.get("/login", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    res.form["password"] = "incorrect horse"
    res = res.form.submit(status=200)
    assert ("error", "Login failed, please check your information") in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Failed password authentication for user",
    ) in caplog.record_tuples


def test_signin_password_substring(testclient, user):
    with testclient.session_transaction() as session:
        assert not session.get("sessions")

    res = testclient.get("/login", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    res.form["password"] = "c"
    res = res.form.submit(status=200)
    assert ("error", "Login failed, please check your information") in res.flashes


def test_signin_bad_csrf(testclient, user):
    with testclient.session_transaction() as session:
        assert not session.get("sessions")

    res = testclient.get("/login", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    res.form["password"] = ""
    res = res.form.submit(status=200)
    assert ("error", "Login failed, please check your information") in res.flashes


def test_signin_with_alternate_attribute(testclient, user):
    res = testclient.get("/login", status=200)

    res.form["login"] = "john@doe.test"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    res.form["password"] = "correct horse battery staple"
    res = res.form.submit()
    res = res.follow(status=302)
    res = res.follow(status=200)

    with testclient.session_transaction() as session:
        assert [{"user": user.id, "last_login_datetime": mock.ANY}] == session.get(
            "sessions"
        )


def test_password_page_without_signin_in_redirects_to_login_page(testclient, user):
    res = testclient.get("/auth/password", status=302)
    assert res.location == "/login"
    assert res.flashes == [
        ("warning", "Cannot remember the login you attempted to sign in with.")
    ]



import json
import unittest.mock
from unittest import mock

from canaille.backends import Backend
from canaille.commands import cli


def test_set_string_by_id(cli_runner, backend, user):
    """Set a string attribute to a model identifier by its id."""
    res = cli_runner.invoke(cli, ["set", "user", user.id, "--given-name", "foobar"])
    assert res.exit_code == 0, res.stdout
    assert json.loads(res.stdout) == {
        "created": mock.ANY,
        "display_name": "Johnny",
        "emails": [
            "john@doe.test",
        ],
        "family_name": "Doe",
        "formatted_address": "1234, some street, 6789 some city, some state",
        "formatted_name": "John (johnny) Doe",
        "given_name": "foobar",
        "id": user.id,
        "last_modified": mock.ANY,
        "password": mock.ANY,
        "phone_numbers": [
            "555-000-000",
        ],
        "preferred_language": "en",
        "profile_url": "https://john.test",
        "user_name": "user",
        "locality": "some city",
        "postal_code": "6789",
        "region": "some state",
        "street": "1234, some street",
        "hotp_counter": 1,
        "last_otp_login": mock.ANY,
        "secret_token": mock.ANY,
    }
    backend.reload(user)
    assert user.given_name == "foobar"


def test_set_string_by_identifier(cli_runner, backend, user):
    """Set a string attribute to a model identifier by its identifier."""
    res = cli_runner.invoke(cli, ["set", "user", "user", "--given-name", "foobar"])
    assert res.exit_code == 0, res.stdout
    assert json.loads(res.stdout) == {
        "created": mock.ANY,
        "display_name": "Johnny",
        "emails": [
            "john@doe.test",
        ],
        "family_name": "Doe",
        "formatted_address": "1234, some street, 6789 some city, some state",
        "formatted_name": "John (johnny) Doe",
        "given_name": "foobar",
        "id": user.id,
        "last_modified": mock.ANY,
        "password": mock.ANY,
        "phone_numbers": [
            "555-000-000",
        ],
        "preferred_language": "en",
        "profile_url": "https://john.test",
        "user_name": "user",
        "locality": "some city",
        "postal_code": "6789",
        "region": "some state",
        "street": "1234, some street",
        "hotp_counter": 1,
        "last_otp_login": mock.ANY,
        "secret_token": mock.ANY,
    }
    backend.reload(user)
    assert user.given_name == "foobar"


def test_set_multiple(cli_runner, backend, user):
    """Test setting several emails."""
    res = cli_runner.invoke(
        cli,
        [
            "set",
            "user",
            user.id,
            "--emails",
            "foo@example.test",
            "--emails",
            "bar@example.test",
        ],
    )
    assert res.exit_code == 0, res.stdout
    assert json.loads(res.stdout) == {
        "created": mock.ANY,
        "display_name": "Johnny",
        "emails": [
            "foo@example.test",
            "bar@example.test",
        ],
        "family_name": "Doe",
        "formatted_address": "1234, some street, 6789 some city, some state",
        "formatted_name": "John (johnny) Doe",
        "given_name": "John",
        "id": user.id,
        "last_modified": mock.ANY,
        "password": mock.ANY,
        "phone_numbers": [
            "555-000-000",
        ],
        "preferred_language": "en",
        "profile_url": "https://john.test",
        "user_name": "user",
        "locality": "some city",
        "postal_code": "6789",
        "region": "some state",
        "street": "1234, some street",
        "hotp_counter": 1,
        "last_otp_login": mock.ANY,
        "secret_token": mock.ANY,
    }
    backend.reload(user)
    assert user.emails == [
        "foo@example.test",
        "bar@example.test",
    ]


def test_set_unknown_id(cli_runner, backend):
    """Error case for trying to set a value for an invalid object."""
    res = cli_runner.invoke(cli, ["set", "user", "invalid", "--given-name", "foobar"])
    assert res.exit_code == 1, res.stdout
    assert res.stderr == "Error: No user with id 'invalid'\n"


def test_set_remove_simple_attribute(cli_runner, backend, user, admin):
    """Test to remove a non multiple attribute."""
    assert user.formatted_address is not None

    res = cli_runner.invoke(cli, ["set", "user", user.id, "--formatted-address", ""])
    assert res.exit_code == 0, res.stdout
    assert json.loads(res.stdout) == {
        "created": mock.ANY,
        "display_name": "Johnny",
        "emails": [
            "john@doe.test",
        ],
        "family_name": "Doe",
        "formatted_name": "John (johnny) Doe",
        "given_name": "John",
        "id": user.id,
        "last_modified": mock.ANY,
        "password": mock.ANY,
        "phone_numbers": [
            "555-000-000",
        ],
        "preferred_language": "en",
        "profile_url": "https://john.test",
        "user_name": "user",
        "locality": "some city",
        "postal_code": "6789",
        "region": "some state",
        "street": "1234, some street",
        "hotp_counter": 1,
        "last_otp_login": mock.ANY,
        "secret_token": mock.ANY,
    }
    backend.reload(user)
    assert user.formatted_address is None


def test_set_remove_multiple_attribute(cli_runner, backend, user, admin, foo_group):
    """Test to remove a non multiple attribute."""
    foo_group.members = [user, admin]
    backend.save(foo_group)
    assert user.groups == [foo_group]

    res = cli_runner.invoke(cli, ["set", "user", user.id, "--groups", ""])
    assert res.exit_code == 0, res.stdout
    assert json.loads(res.stdout) == {
        "created": mock.ANY,
        "display_name": "Johnny",
        "emails": [
            "john@doe.test",
        ],
        "family_name": "Doe",
        "formatted_name": "John (johnny) Doe",
        "formatted_address": "1234, some street, 6789 some city, some state",
        "given_name": "John",
        "id": user.id,
        "last_modified": mock.ANY,
        "password": mock.ANY,
        "phone_numbers": [
            "555-000-000",
        ],
        "preferred_language": "en",
        "profile_url": "https://john.test",
        "user_name": "user",
        "locality": "some city",
        "postal_code": "6789",
        "region": "some state",
        "street": "1234, some street",
        "hotp_counter": 1,
        "last_otp_login": mock.ANY,
        "secret_token": mock.ANY,
    }
    backend.reload(user)
    assert user.groups == []


def test_set_quiet(cli_runner, backend, user):
    """Test that --quiet suppresses output."""
    res = cli_runner.invoke(
        cli, ["set", "user", user.id, "--given-name", "foobar", "--quiet"]
    )
    assert res.exit_code == 0, res.stdout
    assert res.stdout == ""
    backend.reload(user)
    assert user.given_name == "foobar"


def test_set_ignore_errors_unknown_id(cli_runner, backend):
    """Test that errors are ignored with --ignore-errors for unknown id."""
    res = cli_runner.invoke(
        cli, ["set", "user", "invalid", "--given-name", "foobar", "--ignore-errors"]
    )
    assert res.exit_code == 0
    assert res.stdout == ""


def test_set_ignore_errors_save_failure(cli_runner, backend, user):
    """Test that save errors are ignored with --ignore-errors."""

    def mock_save(*args, **kwargs):
        raise Exception("Simulated save error")

    with unittest.mock.patch.object(Backend.instance, "save", side_effect=mock_save):
        res = cli_runner.invoke(cli, ["set", "user", user.id, "--given-name", "foobar"])
        assert res.exit_code != 0
        assert "Simulated save error" in str(res.output)

    with unittest.mock.patch.object(Backend.instance, "save", side_effect=mock_save):
        res = cli_runner.invoke(
            cli, ["set", "user", user.id, "--given-name", "foobar", "--ignore-errors"]
        )
        assert res.exit_code == 0
        assert res.stdout == ""



import datetime

from flask import Blueprint
from flask import current_app
from flask import flash
from flask import g
from flask import redirect
from flask import request
from flask import url_for

from canaille.app.i18n import gettext as _
from canaille.app.templating import render_template
from canaille.backends import Backend
from canaille.core.auth import auth_step
from canaille.core.auth import redirect_to_next_auth_step
from canaille.core.endpoints.forms import TwoFactorForm
from canaille.core.mails import send_one_time_password_mail
from canaille.core.models import SEND_NEW_OTP_DELAY

bp = Blueprint("email", __name__)


@bp.context_processor
def global_processor():
    return {
        "menu": False,
    }


@bp.route("/auth/email", methods=["GET", "POST"])
@auth_step("email")
def email():
    # Automatically send a passcode when users access the form for the first time
    if g.auth.user and (
        not g.auth.user.one_time_password_emission_date
        or g.auth.current_step_start_dt > g.auth.user.one_time_password_emission_date
    ):
        send_email_otp(flashes=False)

    form = TwoFactorForm(request.form or None)
    form.render_field_macro_file = "core/partial/login_field.html"
    resend_delay = (
        g.auth.user.next_otp_send_delay
        if g.auth.user
        else datetime.timedelta(seconds=SEND_NEW_OTP_DELAY)
    )

    if not request.form or form.form_control():
        return render_template(
            "core/auth/email.html", form=form, resend_delay=resend_delay
        )

    if request.form.get("action") == "resend":
        send_email_otp()
        return redirect(url_for(".email"))

    if (
        not form.validate()
        or not g.auth.user
        or not g.auth.user.is_email_or_sms_otp_valid(form.otp.data)
    ):
        flash(_("The passcode you entered is invalid. Please try again"), "error")
        current_app.logger.security(
            f"Failed email code authentication for {g.auth.user_name}"
        )
        return render_template(
            "core/auth/email.html", form=form, resend_delay=resend_delay
        )

    current_app.logger.security(
        f"Successful email code authentication for {g.auth.user_name}"
    )
    g.auth.set_step_finished("email")
    return redirect_to_next_auth_step()


def send_email_otp(flashes=True):
    if g.auth.user:
        if not g.auth.user.can_send_new_otp():
            if flashes:
                flash(
                    _(
                        f"Too many attempts. Please try again in {SEND_NEW_OTP_DELAY} seconds."
                    ),
                    "danger",
                )
            return

        otp = g.auth.user.generate_sms_or_mail_otp()
        if send_one_time_password_mail(g.auth.user.preferred_email, otp):
            Backend.instance.save(g.auth.user)
            current_app.logger.security(
                f"Sent one-time passcode for {g.auth.user_name} to {g.auth.user.preferred_email}"
            )

    if flashes:
        flash(_("The new verification code have been sent."), "success")



import datetime
from dataclasses import dataclass
from functools import wraps

from flask import current_app
from flask import flash
from flask import g
from flask import redirect
from flask import session
from flask import url_for

from canaille.app.i18n import gettext as _
from canaille.app.session import is_user_in_login_history
from canaille.app.session import login_user
from canaille.app.session import save_user_session
from canaille.backends import Backend
from canaille.core.models import User

AUTHENTICATION_FACTORS = {}


def auth_step(step_name=None):
    """Decorate authentication steps and perform basic checks."""

    def wrapper(view_function):
        AUTHENTICATION_FACTORS[step_name] = view_function

        @wraps(view_function)
        def decorator(*args, **kwargs):
            if not g.auth:
                if g.session:
                    return redirect(url_for("core.account.index"))

                flash(
                    _("Cannot remember the login you attempted to sign in with."),
                    "warning",
                )
                return redirect(url_for("core.auth.login"))

            if (
                isinstance(g.auth.current_step, str)
                and step_name != g.auth.current_step
            ) or (
                isinstance(g.auth.current_step, list)
                and step_name not in g.auth.current_step
            ):
                return redirect_to_next_auth_step()

            return view_function(*args, **kwargs)

        return decorator

    return wrapper


@dataclass
class AuthenticationSession:
    user_name: str | None
    """The user name attempting to log-in."""

    remaining: list[str]
    """The remaining auth factors to succeed before logging in."""

    achieved: list[str]
    """The achieved auth factors."""

    current_step_start_dt: datetime.datetime | None
    """The time when the current step has been started."""

    current_step_try_dt: datetime.datetime | None
    """The time when the current try of the current step has started."""

    welcome_flash: bool
    """Display a welcom flash when users are logged in."""

    data: dict[str, str]
    """Custom auth factor data."""

    template: str | None = None
    """Path to base template for authentication pages."""

    known_user: bool = False
    """Whether the user has already logged-in on this device."""

    remember: bool = True
    """Whether to create permanent session and add to login history."""

    _user: User | None = None

    def __init__(
        self,
        user_name=None,
        remaining=None,
        achieved=None,
        current_step_start_dt=None,
        current_step_try_dt=None,
        welcome_flash=None,
        data=None,
        template=None,
        known_user=None,
        remember=None,
    ):
        self.user_name = user_name
        self.data = data or {}
        self.template = template or "core/auth/base.html"
        self.welcome_flash = welcome_flash if welcome_flash is not None else True
        self.known_user = known_user or is_user_in_login_history(user_name)
        self.remember = remember if remember is not None else True
        self.reset_auth_steps(
            remaining, achieved, current_step_start_dt, current_step_try_dt
        )

    def reset_auth_steps(
        self,
        remaining=None,
        achieved=None,
        current_step_start_dt=None,
        current_step_try_dt=None,
    ):
        self.remaining = (
            remaining or current_app.config["CANAILLE"]["AUTHENTICATION_FACTORS"]
        )
        self.achieved = achieved or []
        self.current_step_start_dt = current_step_start_dt or None
        self.current_step_try_dt = current_step_try_dt or None

    def serialize(self):
        payload = {
            "user_name": self.user_name,
            "welcome_flash": self.welcome_flash,
            "remaining": self.remaining,
            "known_user": self.known_user,
            "remember": self.remember,
            "template": self.template,
        }

        if self.current_step_start_dt:
            payload["current_step_start_dt"] = self.current_step_start_dt.isoformat()

        if self.current_step_try_dt:
            payload["current_step_try_dt"] = self.current_step_try_dt.isoformat()

        if self.achieved:
            payload["achieved"] = self.achieved

        if self.data:
            payload["data"] = self.data

        return payload

    @classmethod
    def deserialize(cls, payload):
        if "current_step_start_dt" in payload:
            payload["current_step_start_dt"] = datetime.datetime.fromisoformat(
                payload["current_step_start_dt"]
            )

        if "current_step_try_dt" in payload:
            payload["current_step_try_dt"] = datetime.datetime.fromisoformat(
                payload["current_step_try_dt"]
            )

        return cls(**payload)

    @classmethod
    def load(cls):
        if "auth" in session:
            return AuthenticationSession.deserialize(session.pop("auth"))

    @classmethod
    def update(cls, **kwargs):
        """Update existing AuthenticationSession from g.auth or create a new one with given parameters."""
        if g.auth:
            # Update existing session with new parameters
            for key, value in kwargs.items():
                setattr(g.auth, key, value)
            return g.auth
        else:
            # Create new session
            return cls(**kwargs)

    def save(self):
        session["auth"] = self.serialize()

    @property
    def current_step(self):
        return self.remaining[0] if self.remaining else None

    @property
    def user(self):
        if not self._user:
            self._user = (
                g.session.user if g.session else get_user_from_login(self.user_name)
            )
        return self._user

    def set_step_started(self):
        self.current_step_start_dt = datetime.datetime.now(datetime.timezone.utc)
        self.current_step_try_dt = self.current_step_start_dt

    def set_step_finished(self, step):
        self.achieved.append(step)
        self.remaining.pop(0)


def redirect_to_next_auth_step():
    if not g.auth.remaining:
        if g.session:
            g.session.last_login_datetime = datetime.datetime.now(datetime.timezone.utc)
            save_user_session()
        else:
            login_user(g.auth.user, remember=g.auth.remember)

        redirection = session.pop("redirect-after-login", None)
        if g.auth.welcome_flash and not redirection:
            flash(
                _("Connection successful. Welcome %(user)s", user=g.session.user.name),
                "success",
            )

        current_app.logger.security(f"Successful authentication for {g.auth.user_name}")

        del g.auth
        return redirect(redirection or url_for("core.account.index"))

    try:
        func = AUTHENTICATION_FACTORS[g.auth.current_step]
    except KeyError:
        del g.auth
        flash(
            _(
                "An error happened during your authentication process. Please try again."
            ),
            "error",
        )
        return redirect(url_for("core.auth.login"))

    g.auth.set_step_started()
    return redirect(url_for(f"core.auth.{g.auth.current_step}.{func.__name__}"))


def get_user_from_login(login=None):
    from canaille.app import models

    login_attributes = current_app.config["CANAILLE"]["LOGIN_ATTRIBUTES"]

    if isinstance(login_attributes, list):
        for attr_name in login_attributes:
            if user := Backend.instance.get(models.User, **{attr_name: login}):
                return user

    if isinstance(login_attributes, dict):
        for attr_name, template in login_attributes.items():
            login_value = current_app.jinja_env.from_string(template).render(
                login=login
            )
            if user := Backend.instance.get(models.User, **{attr_name: login_value}):
                return user

    return None


def login_placeholder():
    """Build a placeholder string for the login form, based on the attributes users can use to identify themselves."""
    default_values = {
        "formatted_name": _("John Doe"),
        "user_name": _("jdoe"),
        "emails": _("john.doe@example.com"),
    }
    placeholders = [
        default_values[attr_name]
        for attr_name in current_app.config["CANAILLE"]["LOGIN_ATTRIBUTES"]
        if default_values.get(attr_name)
    ]
    return _(" or ").join(placeholders)



import pytest
from scim2_tester import Status
from scim2_tester import check_server
from scim2_tester.discovery import get_all_available_tags
from scim2_tester.discovery import get_standard_resource_types


@pytest.mark.parametrize("tag", get_all_available_tags())
@pytest.mark.parametrize("resource_type", [None] + get_standard_resource_types())
def test_individual_filters(scim_client, tag, resource_type):
    results = check_server(
        scim_client,
        raise_exceptions=True,
        include_tags={tag},
        resource_types=resource_type,
    )

    for result in results:
        assert result.status in (Status.SKIPPED, Status.SUCCESS)



import json

from canaille.scim.endpoints import bp


def test_patch_user_replace_simple_attribute(
    app, backend, oidc_token, user, scim_client
):
    """Test PATCH operation to replace a simple user attribute."""
    assert user.given_name == "John"
    assert backend.check_user_password(user, "correct horse battery staple")[0]

    patch_data = {
        "schemas": ["urn:ietf:params:scim:api:messages:2.0:PatchOp"],
        "Operations": [{"op": "replace", "path": "title", "value": "Senior Developer"}],
    }
    response = scim_client.client.patch(
        f"{bp.url_prefix}/Users/{user.id}",
        json=patch_data,
        headers={"Authorization": f"Bearer {oidc_token.access_token}"},
    )

    assert response.status_code == 200
    response_data = json.loads(response.data)
    assert response_data["title"] == "Senior Developer"

    backend.reload(user)
    assert user.title == "Senior Developer"
    assert user.given_name == "John"
    assert backend.check_user_password(user, "correct horse battery staple")[0]


def test_patch_user_activate_deactivate(app, backend, oidc_token, user, scim_client):
    """Test PATCH operation to replace a boolean attribute."""
    assert user.given_name == "John"
    assert backend.check_user_password(user, "correct horse battery staple")[0]

    patch_data = {
        "schemas": ["urn:ietf:params:scim:api:messages:2.0:PatchOp"],
        "Operations": [{"op": "replace", "path": "active", "value": False}],
    }
    response = scim_client.client.patch(
        f"{bp.url_prefix}/Users/{user.id}",
        json=patch_data,
        headers={"Authorization": f"Bearer {oidc_token.access_token}"},
    )

    assert response.status_code == 200
    response_data = json.loads(response.data)
    assert response_data["active"] is False

    backend.reload(user)
    assert user.lock_date
    assert user.given_name == "John"

    patch_data = {
        "schemas": ["urn:ietf:params:scim:api:messages:2.0:PatchOp"],
        "Operations": [{"op": "replace", "path": "active", "value": True}],
    }
    response = scim_client.client.patch(
        f"{bp.url_prefix}/Users/{user.id}",
        json=patch_data,
        headers={"Authorization": f"Bearer {oidc_token.access_token}"},
    )

    assert response.status_code == 200
    response_data = json.loads(response.data)
    assert response_data["active"] is True

    backend.reload(user)
    assert not user.lock_date
    assert user.given_name == "John"
    assert backend.check_user_password(user, "correct horse battery staple")[0]


def test_patch_replace_user_password(app, backend, oidc_token, user, scim_client):
    """Test PATCH operation to replace user passwords."""
    assert backend.check_user_password(user, "correct horse battery staple")[0]

    patch_data = {
        "schemas": ["urn:ietf:params:scim:api:messages:2.0:PatchOp"],
        "Operations": [
            {
                "op": "replace",
                "path": "password",
                "value": "incorrect horse battery staple",
            }
        ],
    }
    response = scim_client.client.patch(
        f"{bp.url_prefix}/Users/{user.id}",
        json=patch_data,
        headers={"Authorization": f"Bearer {oidc_token.access_token}"},
    )

    assert response.status_code == 200

    backend.reload(user)
    assert backend.check_user_password(user, "incorrect horse battery staple")[0]


def test_patch_delete_user_password(app, backend, oidc_token, user, scim_client):
    """Test PATCH operation to delete user passwords."""
    assert backend.check_user_password(user, "correct horse battery staple")[0]

    patch_data = {
        "schemas": ["urn:ietf:params:scim:api:messages:2.0:PatchOp"],
        "Operations": [
            {
                "op": "remove",
                "path": "password",
            }
        ],
    }
    response = scim_client.client.patch(
        f"{bp.url_prefix}/Users/{user.id}",
        json=patch_data,
        headers={"Authorization": f"Bearer {oidc_token.access_token}"},
    )

    assert response.status_code == 200
    backend.reload(user)
    assert not backend.check_user_password(user, "correct horse battery staple")[0]


def test_patch_user_add_email(app, backend, oidc_token, user, scim_client):
    """Test PATCH operation to add an email to a user."""
    patch_data = {
        "schemas": ["urn:ietf:params:scim:api:messages:2.0:PatchOp"],
        "Operations": [
            {
                "op": "add",
                "path": "emails",
                "value": [{"value": "newemail@example.com"}],
            }
        ],
    }

    response = scim_client.client.patch(
        f"{bp.url_prefix}/Users/{user.id}",
        json=patch_data,
        headers={"Authorization": f"Bearer {oidc_token.access_token}"},
    )

    assert response.status_code == 200
    response_data = json.loads(response.data)

    emails = response_data.get("emails", [])
    email_values = [email["value"] for email in emails]
    assert "newemail@example.com" in email_values

    backend.reload(user)
    assert "newemail@example.com" in user.emails


def test_patch_user_remove_attribute(app, backend, oidc_token, user, scim_client):
    """Test PATCH operation to remove an attribute from a user."""
    user.title = "Manager"
    backend.save(user)

    patch_data = {
        "schemas": ["urn:ietf:params:scim:api:messages:2.0:PatchOp"],
        "Operations": [{"op": "remove", "path": "title"}],
    }

    response = scim_client.client.patch(
        f"{bp.url_prefix}/Users/{user.id}",
        json=patch_data,
        headers={"Authorization": f"Bearer {oidc_token.access_token}"},
    )

    assert response.status_code == 200
    response_data = json.loads(response.data)
    assert "title" not in response_data or response_data["title"] is None

    backend.reload(user)
    assert user.title is None


def test_patch_user_multiple_operations(app, backend, oidc_token, user, scim_client):
    """Test PATCH with multiple operations in a single request."""
    patch_data = {
        "schemas": ["urn:ietf:params:scim:api:messages:2.0:PatchOp"],
        "Operations": [
            {"op": "replace", "path": "title", "value": "CTO"},
            {"op": "add", "path": "preferredLanguage", "value": "en-US"},
        ],
    }

    response = scim_client.client.patch(
        f"{bp.url_prefix}/Users/{user.id}",
        json=patch_data,
        headers={"Authorization": f"Bearer {oidc_token.access_token}"},
    )

    assert response.status_code == 200
    response_data = json.loads(response.data)
    assert response_data["title"] == "CTO"
    assert response_data["preferredLanguage"] == "en-US"

    backend.reload(user)
    assert user.title == "CTO"
    assert user.preferred_language == "en-US"


def test_patch_group_replace_members(
    app, backend, oidc_token, foo_group, user, moderator, scim_client
):
    """Test PATCH operation to replace members in a group."""
    foo_group.members = [user]
    backend.save(foo_group)

    patch_data = {
        "schemas": ["urn:ietf:params:scim:api:messages:2.0:PatchOp"],
        "Operations": [
            {
                "op": "replace",
                "path": "members",
                "value": [
                    {
                        "value": moderator.id,
                        "display": moderator.display_name,
                    }
                ],
            }
        ],
    }

    response = scim_client.client.patch(
        f"{bp.url_prefix}/Groups/{foo_group.id}",
        json=patch_data,
        headers={"Authorization": f"Bearer {oidc_token.access_token}"},
    )

    assert response.status_code == 200
    response_data = json.loads(response.data)

    members = response_data.get("members", [])
    member_ids = [member["value"] for member in members]
    assert moderator.identifier in member_ids
    assert user.id not in member_ids

    backend.reload(foo_group)
    member_ids_persisted = [member.identifier for member in foo_group.members]
    assert moderator.identifier in member_ids_persisted
    assert user.identifier not in member_ids_persisted


def test_patch_group_add_invalid_member(
    app, backend, oidc_token, foo_group, user, scim_client
):
    """Test PATCH operation to add a member to a group."""
    foo_group.members = [user]
    backend.save(foo_group)

    patch_data = {
        "schemas": ["urn:ietf:params:scim:api:messages:2.0:PatchOp"],
        "Operations": [
            {
                "op": "add",
                "path": "members",
                "value": [
                    {
                        "value": "invalid",
                        "display": "http://invalid",
                    }
                ],
            }
        ],
    }

    response = scim_client.client.patch(
        f"{bp.url_prefix}/Groups/{foo_group.id}",
        json=patch_data,
        headers={"Authorization": f"Bearer {oidc_token.access_token}"},
    )

    assert response.status_code == 200
    response_data = json.loads(response.data)

    members = response_data.get("members", [])
    member_ids = [member["value"] for member in members]
    assert member_ids == [user.identifier]


def test_patch_group_add_member(
    app, backend, oidc_token, foo_group, user, moderator, scim_client
):
    """Test PATCH operation to add a member to a group."""
    foo_group.members = [user]
    backend.save(foo_group)

    patch_data = {
        "schemas": ["urn:ietf:params:scim:api:messages:2.0:PatchOp"],
        "Operations": [
            {
                "op": "add",
                "path": "members",
                "value": [
                    {
                        "value": moderator.id,
                        "display": moderator.display_name,
                    }
                ],
            }
        ],
    }

    response = scim_client.client.patch(
        f"{bp.url_prefix}/Groups/{foo_group.id}",
        json=patch_data,
        headers={"Authorization": f"Bearer {oidc_token.access_token}"},
    )

    assert response.status_code == 200
    response_data = json.loads(response.data)

    members = response_data.get("members", [])
    member_ids = [member["value"] for member in members]
    assert user.identifier in member_ids
    assert moderator.identifier in member_ids

    backend.reload(foo_group)
    member_ids_persisted = [member.identifier for member in foo_group.members]
    assert user.identifier in member_ids_persisted
    assert moderator.identifier in member_ids_persisted


def test_patch_invalid_operation(app, backend, oidc_token, user, scim_client):
    """Test PATCH with an invalid operation type."""
    patch_data = {
        "schemas": ["urn:ietf:params:scim:api:messages:2.0:PatchOp"],
        "Operations": [{"op": "invalid_operation", "path": "title", "value": "Test"}],
    }

    response = scim_client.client.patch(
        f"{bp.url_prefix}/Users/{user.id}",
        json=patch_data,
        headers={"Authorization": f"Bearer {oidc_token.access_token}"},
    )

    assert response.status_code == 400


def test_patch_user_no_modifications(app, backend, oidc_token, user, scim_client):
    """Test PATCH user that results in no modifications."""
    user.title = "Developer"
    backend.save(user)

    patch_data = {
        "schemas": ["urn:ietf:params:scim:api:messages:2.0:PatchOp"],
        "Operations": [
            {
                "op": "replace",
                "path": "title",
                "value": "Developer",  # Same as current value
            }
        ],
    }

    response = scim_client.client.patch(
        f"{bp.url_prefix}/Users/{user.id}",
        json=patch_data,
        headers={"Authorization": f"Bearer {oidc_token.access_token}"},
    )

    assert response.status_code == 200
    response_data = json.loads(response.data)
    assert response_data["title"] == "Developer"

    backend.reload(user)
    assert user.title == "Developer"


def test_patch_group_no_modification(
    app, backend, oidc_token, foo_group, user, scim_client
):
    """Test PATCH user that results in no modifications."""
    foo_group.members = [user]
    backend.save(foo_group)

    patch_data = {
        "schemas": ["urn:ietf:params:scim:api:messages:2.0:PatchOp"],
        "Operations": [
            {
                "op": "replace",
                "path": "displayName",
                "value": "foo",
            }
        ],
    }

    response = scim_client.client.patch(
        f"{bp.url_prefix}/Groups/{foo_group.id}",
        json=patch_data,
        headers={"Authorization": f"Bearer {oidc_token.access_token}"},
    )

    assert response.status_code == 200
    response_data = json.loads(response.data)

    members = response_data.get("members", [])
    member_ids = [member["value"] for member in members]
    assert user.identifier in member_ids



import datetime
import logging
import time
import uuid
from urllib.parse import parse_qs
from urllib.parse import urlsplit

from joserfc import jwt

from canaille.app import models

from . import client_credentials


def test_refresh_token(testclient, logged_user, client, backend, caplog):
    assert not backend.query(models.Consent)

    res = testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="code",
            client_id=client.client_id,
            scope="openid profile",
            nonce="somenonce",
            redirect_uri="https://client.test/redirect1",
        ),
    )
    res = res.form.submit(name="answer", value="accept")

    assert res.location.startswith(client.redirect_uris[0])
    params = parse_qs(urlsplit(res.location).query)
    code = params["code"][0]
    authcode = backend.get(models.AuthorizationCode, code=code)
    assert authcode is not None

    consents = backend.query(models.Consent, client=client, subject=logged_user)
    assert "profile" in consents[0].scope

    res = testclient.post(
        "/oauth/token",
        params=dict(
            grant_type="authorization_code",
            code=code,
            scope="openid profile",
            redirect_uri=client.redirect_uris[0],
        ),
        headers={"Authorization": f"Basic {client_credentials(client)}"},
        status=200,
    )
    access_token = res.json["access_token"]
    old_token = backend.get(models.Token, access_token=access_token)
    assert old_token is not None
    assert not old_token.revokation_date

    res = testclient.post(
        "/oauth/token",
        params=dict(
            grant_type="refresh_token",
            refresh_token=res.json["refresh_token"],
        ),
        headers={"Authorization": f"Basic {client_credentials(client)}"},
        status=200,
    )
    access_token = res.json["access_token"]
    new_token = backend.get(models.Token, access_token=access_token)
    assert new_token is not None
    assert old_token.access_token != new_token.access_token
    assert (
        "canaille",
        logging.SECURITY,
        "Issued refresh_token token for user in client Some client",
    ) in caplog.record_tuples
    backend.reload(old_token)
    assert old_token.revokation_date

    res = testclient.get(
        "/oauth/userinfo",
        headers={"Authorization": f"Bearer {access_token}"},
        status=200,
    )
    assert res.json["name"] == "John (johnny) Doe"

    for consent in consents:
        backend.delete(consent)


def test_refresh_token_with_invalid_user(testclient, client, backend):
    user = models.User(
        formatted_name="John Doe",
        family_name="Doe",
        user_name="temp",
        emails=["temp@temp.test"],
        password="correct horse battery staple",
    )
    backend.save(user)

    res = testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="code",
            client_id=client.client_id,
            scope="openid profile",
            nonce="somenonce",
            redirect_uri="https://client.test/redirect1",
        ),
    ).follow()

    res.form["login"] = "temp"
    res = res.form.submit(name="answer", value="accept", status=302).follow()

    res.form["password"] = "correct horse battery staple"
    res = res.form.submit(name="answer", value="accept", status=302).follow()
    res = res.form.submit(name="answer", value="accept", status=302)

    params = parse_qs(urlsplit(res.location).query)
    code = params["code"][0]
    backend.get(models.AuthorizationCode, code=code)

    res = testclient.post(
        "/oauth/token",
        params=dict(
            grant_type="authorization_code",
            code=code,
            scope="openid profile",
            redirect_uri=client.redirect_uris[0],
        ),
        headers={"Authorization": f"Basic {client_credentials(client)}"},
        status=200,
    )

    refresh_token = res.json["refresh_token"]
    access_token = res.json["access_token"]
    backend.delete(user)

    res = testclient.post(
        "/oauth/token",
        params=dict(
            grant_type="refresh_token",
            refresh_token=refresh_token,
        ),
        headers={"Authorization": f"Basic {client_credentials(client)}"},
        status=400,
    )
    assert res.json == {
        "error": "invalid_request",
        "error_description": "There is no 'user' for this token.",
    }
    token = backend.get(models.Token, access_token=access_token)
    backend.delete(token)


def test_cannot_refresh_token_for_locked_users(
    testclient, logged_user, client, backend
):
    """Canaille should not issue new tokens for locked users."""
    res = testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="code",
            client_id=client.client_id,
            scope="openid profile",
            nonce="somenonce",
            redirect_uri="https://client.test/redirect1",
        ),
    )
    res = res.form.submit(name="answer", value="accept")

    assert res.location.startswith(client.redirect_uris[0])
    params = parse_qs(urlsplit(res.location).query)
    code = params["code"][0]

    res = testclient.post(
        "/oauth/token",
        params=dict(
            grant_type="authorization_code",
            code=code,
            scope="openid profile",
            redirect_uri=client.redirect_uris[0],
        ),
        headers={"Authorization": f"Basic {client_credentials(client)}"},
        status=200,
    )

    logged_user.lock_date = datetime.datetime.now(datetime.timezone.utc)
    backend.save(logged_user)

    res = testclient.post(
        "/oauth/token",
        params=dict(
            grant_type="refresh_token",
            refresh_token=res.json["refresh_token"],
        ),
        headers={"Authorization": f"Basic {client_credentials(client)}"},
        status=400,
    )
    assert "access_token" not in res.json


def test_jwt_auth(testclient, client_jwk, client, token, caplog, backend, user):
    """Test client JWT authentication as defined per RFC7523, for a refresh_token."""
    now = time.time()

    client.client_uri = "https://client.trusted.test"
    client.token_endpoint_auth_method = "client_secret_jwt"
    backend.save(client)

    header = {"alg": "RS256"}
    payload = {
        "iss": client.client_id,
        "sub": client.client_id,
        "aud": "http://canaille.test/oauth/token",
        "nbf": now - 3600,
        "exp": now + 3600,
        "iat": now - 1,
        "jti": str(uuid.uuid4()),
    }
    client_jwt = jwt.encode(header, payload, client_jwk)
    res = testclient.post(
        "/oauth/token",
        params=dict(
            grant_type="refresh_token",
            refresh_token=token.refresh_token,
            client_assertion=client_jwt,
            client_assertion_type="urn:ietf:params:oauth:client-assertion-type:jwt-bearer",
            scope=token.scope,
            redirect_uri=client.redirect_uris[0],
        ),
        status=200,
    )

    access_token = res.json["access_token"]
    token = backend.get(models.Token, access_token=access_token)
    assert token.client == client
    assert token.subject == user



import datetime
import json
import typing
import uuid

from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import LargeBinary
from sqlalchemy import String
from sqlalchemy import Table
from sqlalchemy import or_
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import class_mapper
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import relationship
from sqlalchemy.orm.properties import ColumnProperty
from sqlalchemy_json import MutableJson
from sqlalchemy_utils import PasswordType
from sqlalchemy_utils import force_auto_coercion

import canaille.core.models
import canaille.oidc.models
from canaille.backends import Backend
from canaille.backends.models import BackendModel

from .backend import Base
from .utils import TZDateTime

force_auto_coercion()


class SqlAlchemyModel(BackendModel):
    __mapper_args__ = {
        # avoids warnings on double deletions
        "confirm_deleted_rows": False,
    }

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} {self.identifier_attribute}={self.identifier}>"
        )

    @classmethod
    def attribute_filter(cls, name, value):
        if isinstance(value, list):
            return or_(cls.attribute_filter(name, v) for v in value)

        multiple = typing.get_origin(cls.attributes[name]) is list

        if multiple:
            column = getattr(cls, name)
            column_property = getattr(cls, name).property
            if not isinstance(column_property, ColumnProperty):
                return column.contains(value)

            # This is a JSON column with primitive values - use exact matching
            # to prevent MultipleResultsFound when searching partial strings
            quoted_value = json.dumps(value)  # Properly escape the value
            return column.cast(String).like(f"%{quoted_value}%")

        return getattr(cls, name) == value

    @classmethod
    def is_attr_required(cls, attr_name: str) -> bool:
        if attr_name in class_mapper(cls).relationships and hasattr(
            cls, f"{attr_name}_id"
        ):
            return cls.is_attr_required(f"{attr_name}_id")

        attr = getattr(cls, attr_name)
        return not getattr(attr, "nullable", True)

    @classmethod
    def is_attr_readonly(cls, attr_name: str) -> bool:
        return False


membership_association_table = Table(
    "membership_association_table",
    Base.metadata,
    Column("index", Integer, autoincrement=True),
    Column("user_id", ForeignKey("user.id"), primary_key=True),
    Column("group_id", ForeignKey("group.id"), primary_key=True),
)


class User(canaille.core.models.User, Base, SqlAlchemyModel):
    __tablename__ = "user"

    @staticmethod
    def default_password_arguments(**kwargs):
        return dict(
            schemes=Backend.instance.config["CANAILLE_SQL"]["PASSWORD_SCHEMES"],
            **kwargs,
        )

    id: Mapped[str] = mapped_column(
        String, primary_key=True, default=lambda: str(uuid.uuid4())
    )
    created: Mapped[datetime.datetime] = mapped_column(
        TZDateTime(timezone=True), nullable=True
    )
    last_modified: Mapped[datetime.datetime] = mapped_column(
        TZDateTime(timezone=True), nullable=True
    )
    user_name: Mapped[str] = mapped_column(String, unique=True, nullable=False)
    password: Mapped[str] = mapped_column(
        PasswordType(
            onload=default_password_arguments,
        ),
        nullable=True,
    )

    def get_password_hash(self):
        if self.password is None:
            return None
        return self.password.hash.decode("utf-8")

    password_last_update: Mapped[datetime.datetime] = mapped_column(
        TZDateTime(timezone=True), nullable=True
    )
    _password_failure_timestamps: Mapped[list[str]] = mapped_column(
        MutableJson, nullable=True
    )
    preferred_language: Mapped[str] = mapped_column(String, nullable=True)
    family_name: Mapped[str] = mapped_column(String, nullable=True)
    given_name: Mapped[str] = mapped_column(String, nullable=True)
    formatted_name: Mapped[str] = mapped_column(String, nullable=True)
    display_name: Mapped[str] = mapped_column(String, nullable=True)
    emails: Mapped[list[str]] = mapped_column(MutableJson, nullable=True)
    phone_numbers: Mapped[list[str]] = mapped_column(MutableJson, nullable=True)
    formatted_address: Mapped[str] = mapped_column(String, nullable=True)
    street: Mapped[str] = mapped_column(String, nullable=True)
    postal_code: Mapped[str] = mapped_column(String, nullable=True)
    locality: Mapped[str] = mapped_column(String, nullable=True)
    region: Mapped[str] = mapped_column(String, nullable=True)
    photo: Mapped[bytes] = mapped_column(LargeBinary, nullable=True)
    profile_url: Mapped[str] = mapped_column(String, nullable=True)
    employee_number: Mapped[str] = mapped_column(String, nullable=True)
    department: Mapped[str] = mapped_column(String, nullable=True)
    title: Mapped[str] = mapped_column(String, nullable=True)
    organization: Mapped[str] = mapped_column(String, nullable=True)
    groups: Mapped[list["Group"]] = relationship(
        secondary=membership_association_table,
        back_populates="members",
        order_by=membership_association_table.c.index,
    )
    lock_date: Mapped[datetime.datetime] = mapped_column(
        TZDateTime(timezone=True), nullable=True
    )
    last_otp_login: Mapped[datetime.datetime] = mapped_column(
        TZDateTime(timezone=True), nullable=True
    )
    secret_token: Mapped[str] = mapped_column(String, nullable=True, unique=True)
    hotp_counter: Mapped[int] = mapped_column(Integer, nullable=True)
    one_time_password: Mapped[str] = mapped_column(String, nullable=True)
    one_time_password_emission_date: Mapped[datetime.datetime] = mapped_column(
        TZDateTime(timezone=True), nullable=True
    )

    @property
    def password_failure_timestamps(self):
        if self._password_failure_timestamps:
            return [
                datetime.datetime.fromisoformat(d)
                for d in self._password_failure_timestamps
            ]
        return self._password_failure_timestamps

    @password_failure_timestamps.setter
    def password_failure_timestamps(self, dates_list):
        if dates_list:
            self._password_failure_timestamps = [str(d) for d in dates_list]
        else:
            self._password_failure_timestamps = dates_list


class Group(canaille.core.models.Group, Base, SqlAlchemyModel):
    __tablename__ = "group"

    id: Mapped[str] = mapped_column(
        String, primary_key=True, default=lambda: str(uuid.uuid4())
    )
    created: Mapped[datetime.datetime] = mapped_column(
        TZDateTime(timezone=True), nullable=True
    )
    last_modified: Mapped[datetime.datetime] = mapped_column(
        TZDateTime(timezone=True), nullable=True
    )

    display_name: Mapped[str] = mapped_column(String)
    description: Mapped[str] = mapped_column(String, nullable=True)
    members: Mapped[list["User"]] = relationship(
        secondary=membership_association_table,
        back_populates="groups",
        order_by=membership_association_table.c.index,
    )


client_audience_association_table = Table(
    "client_audience_association_table",
    Base.metadata,
    Column("index", Integer, autoincrement=True),
    Column("audience_id", ForeignKey("client.id"), primary_key=True, nullable=True),
    Column("client_id", ForeignKey("client.id"), primary_key=True, nullable=True),
)


class Client(canaille.oidc.models.Client, Base, SqlAlchemyModel):
    __tablename__ = "client"

    id: Mapped[str] = mapped_column(
        String, primary_key=True, default=lambda: str(uuid.uuid4())
    )
    created: Mapped[datetime.datetime] = mapped_column(
        TZDateTime(timezone=True), nullable=True
    )
    last_modified: Mapped[datetime.datetime] = mapped_column(
        TZDateTime(timezone=True), nullable=True
    )

    description: Mapped[str] = mapped_column(String, nullable=True)
    post_logout_redirect_uris: Mapped[list[str]] = mapped_column(
        MutableJson, nullable=True
    )
    audience: Mapped[list["Client"]] = relationship(
        "Client",
        secondary=client_audience_association_table,
        primaryjoin=id == client_audience_association_table.c.client_id,
        secondaryjoin=id == client_audience_association_table.c.audience_id,
        order_by=client_audience_association_table.c.index,
    )
    client_id: Mapped[str] = mapped_column(String, nullable=True)
    client_secret: Mapped[str] = mapped_column(String, nullable=True)
    client_id_issued_at: Mapped[datetime.datetime] = mapped_column(
        TZDateTime(timezone=True), nullable=True
    )
    client_secret_expires_at: Mapped[datetime.datetime] = mapped_column(
        TZDateTime(timezone=True), nullable=True
    )
    client_name: Mapped[str] = mapped_column(String, nullable=True)
    contacts: Mapped[list[str]] = mapped_column(MutableJson, nullable=True)
    client_uri: Mapped[str] = mapped_column(String, nullable=True)
    redirect_uris: Mapped[list[str]] = mapped_column(MutableJson, nullable=True)
    logo_uri: Mapped[str] = mapped_column(String, nullable=True)
    grant_types: Mapped[list[str]] = mapped_column(MutableJson, nullable=True)
    response_types: Mapped[list[str]] = mapped_column(MutableJson, nullable=True)
    scope: Mapped[list[str]] = mapped_column(MutableJson, nullable=True)
    tos_uri: Mapped[str] = mapped_column(String, nullable=True)
    policy_uri: Mapped[str] = mapped_column(String, nullable=True)
    jwks_uri: Mapped[str] = mapped_column(String, nullable=True)
    jwks: Mapped[str] = mapped_column(String, nullable=True)
    token_endpoint_auth_method: Mapped[str] = mapped_column(String, nullable=True)
    token_endpoint_auth_signing_alg: Mapped[str] = mapped_column(String, nullable=True)
    sector_identifier_uri: Mapped[str] = mapped_column(String, nullable=True)
    subject_type: Mapped[str] = mapped_column(String, nullable=True)
    software_id: Mapped[str] = mapped_column(String, nullable=True)
    software_version: Mapped[str] = mapped_column(String, nullable=True)
    application_type: Mapped[str] = mapped_column(String, nullable=True)
    id_token_signed_response_alg: Mapped[str] = mapped_column(String, nullable=True)
    id_token_encrypted_response_alg: Mapped[str] = mapped_column(String, nullable=True)
    id_token_encrypted_response_enc: Mapped[str] = mapped_column(String, nullable=True)
    userinfo_signed_response_alg: Mapped[str] = mapped_column(String, nullable=True)
    userinfo_encrypted_response_alg: Mapped[str] = mapped_column(String, nullable=True)
    userinfo_encrypted_response_enc: Mapped[str] = mapped_column(String, nullable=True)
    default_max_age: Mapped[int] = mapped_column(Integer, nullable=True)
    require_auth_time: Mapped[bool] = mapped_column(Boolean, nullable=True)
    default_acr_values: Mapped[list[str]] = mapped_column(
        MutableJson, nullable=True, default=[]
    )
    initiate_login_uri: Mapped[str] = mapped_column(String, nullable=True)
    request_object_signing_alg: Mapped[str] = mapped_column(String, nullable=True)
    request_object_encryption_alg: Mapped[str] = mapped_column(String, nullable=True)
    request_object_encryption_enc: Mapped[str] = mapped_column(String, nullable=True)
    request_uris: Mapped[list[str]] = mapped_column(
        MutableJson, nullable=True, default=[]
    )
    require_signed_request_object: Mapped[bool] = mapped_column(Boolean, nullable=True)


class AuthorizationCode(canaille.oidc.models.AuthorizationCode, Base, SqlAlchemyModel):
    __tablename__ = "authorization_code"

    id: Mapped[str] = mapped_column(
        String, primary_key=True, default=lambda: str(uuid.uuid4())
    )
    created: Mapped[datetime.datetime] = mapped_column(
        TZDateTime(timezone=True), nullable=True
    )
    last_modified: Mapped[datetime.datetime] = mapped_column(
        TZDateTime(timezone=True), nullable=True
    )

    authorization_code_id: Mapped[str] = mapped_column(String, nullable=True)
    code: Mapped[str] = mapped_column(String, nullable=True)
    client_id: Mapped[str] = mapped_column(ForeignKey("client.id"))
    client: Mapped["Client"] = relationship()
    subject_id: Mapped[str] = mapped_column(ForeignKey("user.id"))
    subject: Mapped["User"] = relationship()
    redirect_uri: Mapped[str] = mapped_column(String, nullable=True)
    response_type: Mapped[str] = mapped_column(String, nullable=True)
    scope: Mapped[list[str]] = mapped_column(MutableJson, nullable=True)
    nonce: Mapped[str] = mapped_column(String, nullable=True)
    issue_date: Mapped[datetime.datetime] = mapped_column(
        TZDateTime(timezone=True), nullable=True
    )
    lifetime: Mapped[int] = mapped_column(Integer, nullable=True)
    challenge: Mapped[str] = mapped_column(String, nullable=True)
    challenge_method: Mapped[str] = mapped_column(String, nullable=True)
    revokation_date: Mapped[datetime.datetime] = mapped_column(
        TZDateTime(timezone=True), nullable=True
    )
    auth_time: Mapped[datetime.datetime] = mapped_column(
        TZDateTime(timezone=True), nullable=True
    )
    acr: Mapped[str] = mapped_column(String, nullable=True)
    amr: Mapped[list[str]] = mapped_column(MutableJson, nullable=True)


token_audience_association_table = Table(
    "token_audience_association_table",
    Base.metadata,
    Column("index", Integer, autoincrement=True),
    Column("token_id", ForeignKey("token.id"), primary_key=True, nullable=True),
    Column("client_id", ForeignKey("client.id"), primary_key=True, nullable=True),
)


class Token(canaille.oidc.models.Token, Base, SqlAlchemyModel):
    __tablename__ = "token"

    id: Mapped[str] = mapped_column(
        String, primary_key=True, default=lambda: str(uuid.uuid4())
    )
    created: Mapped[datetime.datetime] = mapped_column(
        TZDateTime(timezone=True), nullable=True
    )
    last_modified: Mapped[datetime.datetime] = mapped_column(
        TZDateTime(timezone=True), nullable=True
    )

    token_id: Mapped[str] = mapped_column(String, nullable=True)
    access_token: Mapped[str] = mapped_column(String, nullable=True, unique=True)
    client_id: Mapped[str] = mapped_column(ForeignKey("client.id"))
    client: Mapped["Client"] = relationship()
    subject_id: Mapped[str] = mapped_column(ForeignKey("user.id"), nullable=True)
    subject: Mapped["User"] = relationship()
    type: Mapped[str] = mapped_column(String, nullable=True)
    refresh_token: Mapped[str] = mapped_column(String, nullable=True, unique=True)
    scope: Mapped[list[str]] = mapped_column(MutableJson, nullable=True)
    issue_date: Mapped[datetime.datetime] = mapped_column(
        TZDateTime(timezone=True), nullable=True
    )
    lifetime: Mapped[int] = mapped_column(Integer, nullable=True)
    revokation_date: Mapped[datetime.datetime] = mapped_column(
        TZDateTime(timezone=True), nullable=True
    )
    audience: Mapped[list["Client"]] = relationship(
        "Client",
        secondary=token_audience_association_table,
        primaryjoin=id == token_audience_association_table.c.token_id,
        secondaryjoin=Client.id == token_audience_association_table.c.client_id,
        order_by=token_audience_association_table.c.index,
    )


class Consent(canaille.oidc.models.Consent, Base, SqlAlchemyModel):
    __tablename__ = "consent"

    id: Mapped[str] = mapped_column(
        String, primary_key=True, default=lambda: str(uuid.uuid4())
    )
    created: Mapped[datetime.datetime] = mapped_column(
        TZDateTime(timezone=True), nullable=True
    )
    last_modified: Mapped[datetime.datetime] = mapped_column(
        TZDateTime(timezone=True), nullable=True
    )

    consent_id: Mapped[str] = mapped_column(String, nullable=True)
    subject_id: Mapped[str] = mapped_column(ForeignKey("user.id"))
    subject: Mapped["User"] = relationship()
    client_id: Mapped[str] = mapped_column(ForeignKey("client.id"))
    client: Mapped["Client"] = relationship()
    scope: Mapped[list[str]] = mapped_column(MutableJson, nullable=True)
    issue_date: Mapped[datetime.datetime] = mapped_column(
        TZDateTime(timezone=True), nullable=True
    )
    revokation_date: Mapped[datetime.datetime] = mapped_column(
        TZDateTime(timezone=True), nullable=True
    )
